#! /usr/bin/python

#     file     : projects/PaminaIII/bin/RunIterativeReplanning.py
#     project  : Pamina III simulation suite
#     author   : Marcus Rickert
#     RCS      : $Id: RunIterativeReplanning.py,v 1.6 2005/04/01 15:47:12 mr Exp $
#     purpose  : script to start and control the various parts of the
#                PAMINA simulation suite
#
#     Copyright (C) 2005 :  Marcus Rickert

# TODO:
#
# * only copy feedback files if size > 0
# * check if archive directory exists
# * new option for replanning fraction of 1.0 for first iteration
# * archive most recently used routefile after iteration terminates
# * only keep a certain number of PAMINA route files
# * optionally delete "old" TRANSIMS routefiles


import posix
import time
import os
import re

from l_File import *
from t_AppException import *
from t_OptionHandler import *
from t_Configuration import *

# options of the skript

OPTION_FIRST_ITERATION = "first-iteration=INDEX"
OPTION_LAST_ITERATION = "last-iteration=INDEX"
OPTION_SKIP_PLANNER = "skip-planner"
OPTION_SKIP_ROUTE_CONVERTER = "skip-route-converter"
OPTION_CONFIG_TEMPLATE = "config-template=FILENAME"
OPTION_WORKING_DIRECTORY = "working-directory"
OPTION_DETECT_RESTART_POINT = "detect-restart-point"
OPTION_CHECK_ARCHIVE = "check-archive"
OPTION_REDIRECT_STDOUT = "redirect-stdout"

# configuration items of group "Iteration"
CONFIG_GROUP_ITERATION = "Iteration"
CONFIG_ITERATION_STRING_FORMAT = "StringFormat"
CONFIG_PAMINA_HOME_DIR = "PaminaHomeDir"
CONFIG_ARCHIVE_DIR = "ArchiveDir"

# configuration items of group "Planner"
CONFIG_GROUP_PLANNER = "Planner"
CONFIG_TRANSIMS_MAP_BASE_FILENAME = "TransimsMapBaseFilename"
CONFIG_REPLANNING_FRACTION = "ReplanningFraction"
CONFIG_REDUCTION_FRACTION = "ReductionFraction"
CONFIG_UPDATE_AGE = "UpdateAge"
CONFIG_INITIAL_AGE = "InitialAge"
CONFIG_LINEAR_AGE_DEPENDANT = "LinearAgeDependant"

# configuration items of group "RouteConverter"
CONFIG_GROUP_ROUTE_CONVERTER = "RouteConverter"
CONFIG_TRANSIMS_PLANFILE_TEMPLATE = "Filename"

# configuration items of group "Router"
CONFIG_GROUP_ROUTER = "Router"
CONFIG_PAMINA_PLANFILE_TEMPLATE = "Routeplan"

# configuration items of grouo "TrafficData"

CONFIG_GROUP_TRAFFICDATA = "TrafficData"
CONFIG_EDGE_TIME_FILE_TEMPLATE = "EdgeTimeFilename"
CONFIG_NODE_TIME_FILE_TEMPLATE = "NodeTimeFilename"

DEFAULT_FIRST_ITERATION = 1
DEFAULT_LAST_ITERATION = 999
DEFAULT_CONFIG_TEMPLATE = "IterationConfig"
DEFAULT_CONFIG = "Config"
DEFAULT_ITERATION_STRING_FORMAT = "%03d"
DEFAULT_PAMINA_HOME_DIR = "../.."
DEFAULT_DETECT_RESTART_POINT = 0
DEFAULT_CHECK_ARCHIVE = 0
DEFAULT_LOG_FILE = "log/iteration.__ITERATION__.log"

DEFAULT_REPLANNING_FRACTION = 0.05
DEFAULT_REDUCTION_FRACTION = 0.0
DEFAULT_UPDATE_AGE = 1
DEFAULT_INITIAL_AGE = 0
DEFAULT_LINEAR_AGE_DEPENDANT = 0


EXTENSION_TRANSIMS_LINK_STAT = ".tls"
EXTENSION_COMPLETED_ROUTES_STAT = ".crt"
EXTENSION_TURN_COUNT_STAT = ".tcs"
EXTENSION_ROUTE_BY_ORIGIN_STAT = ".rbo"
EXTENSION_TRANSIMS_LINK_HIT_STAT = ".lhs"
EXTENSION_EDGE_TIME = ".edg.time"
EXTENSION_NODE_TIME = ".nod.time"
EXTENSION_OLD_TIME = ".old"
EXTENSION_NEW_TIME = ".new"
EXTENSION_GZIP = ".gz"
EXTENSION_UNKNOWN_ARCH = ".UNKNOWN_ARCH"
EXTENSION_INDEX = ".idx"

ENV_COMMUNICATION = "COMMUNICATION"
ENV_PVM_ROOT = "PVM_ROOT"
ENV_PVM_ARCH = "PVM_ARCH"
ENV_HOME = "HOME"

VALUE_MPI = "MPI"
VALUE_PVM = "PVM"

PAMINA_EXECUTABLE = "ParSim"
ROUTE_CONVERTER_EXECUTABLE = "RouteConverter"
PLANNER_EXECUTABLE = "Planner"

PLANNER_SUB_DIR = "src/Planner"

# global variables

g_Communication = None
g_HomeDir = None
g_WorkingDir = None
g_PvmRoot = None
g_PvmArch = None
g_PvmLibDir = None
g_PvmBinDir = None
g_FirstIteration = None
g_LastIteration = None
g_ConfigTemplateFile = None
g_ConfigFile = DEFAULT_CONFIG
g_Config = None
g_IterationStringFormat = None
g_RedirectSuffix = ""

g_BinaryMode = 0

g_ReplanningFraction = 0.05
g_ReductionFraction = 0.0
g_UpdateAge = 1
g_InitialAge = None
g_LinearAgeDependant = 0

# -----------------------------------------------------------------------------------

def CheckPvmEnvironment():

    global g_PvmRoot, g_PvmArch, g_PvmLibDir, g_PvmBinDir, g_Arch
    global g_PaminaExecutable, g_RouteConverterExecutable, g_PlannerExecutable
    
    print "Checking PVM environment..."
    if not os.environ.has_key(ENV_PVM_ROOT):
        raise t_AppException(p_Message = "environment variable %s must be set" % ENV_PVM_ROOT)
    g_PvmRoot = os.environ[ENV_PVM_ROOT]
    print "Found %s='%s'" % (ENV_PVM_ROOT, g_PvmRoot)
    
    g_PvmArch = os.environ[ENV_PVM_ARCH]
    if g_PvmArch == None:
        raise t_AppException(p_Message = "environment variable %s must be set" % ENV_PVM_ARCH)
    print "Found %s='%s'" % (ENV_PVM_ARCH, g_PvmArch)

    g_PvmLibDir = os.path.join(g_PvmRoot, "lib", g_PvmArch)

    if not os.path.isdir(g_PvmLibDir):
        raise t_AppException(p_Message = "PVM not installed in %s" % g_PvmLibDir)

    print "Found PVM library directory %s" % g_PvmLibDir

    g_PvmBinDir = os.path.join(g_HomeDir + "/", "pvm3/bin", g_PvmArch)

    if not os.path.isdir(g_PvmBinDir):
        raise t_AppException(p_Message = "no user PVM directory found in %s" % g_PvmBinDir)

    print "Found user PVM bin directory %s" % g_PvmBinDir

    g_PaminaExecutable = os.path.join(g_PvmBinDir, PAMINA_EXECUTABLE)
    g_RouteConverterExecutable = os.path.join(g_PvmBinDir, ROUTE_CONVERTER_EXECUTABLE)
    g_PlannerExecutable = os.path.join(g_PvmBinDir, PLANNER_EXECUTABLE)
    g_Arch = g_PvmArch


# -----------------------------------------------------------------------------------

def CheckMpiEnvironment():
    print "Checking MPI environment..."

# -----------------------------------------------------------------------------------

def CheckEnvironment():

    global g_HomeDir, g_Communication

    print "Checking environment..."

    if not os.environ.has_key(ENV_HOME):
        raise t_AppException(p_Message = ("environment variable %s must be set" % ENV_HOME))

    g_HomeDir = os.environ[ENV_HOME]
    
    print "Found %s='%s'" % (ENV_HOME, g_HomeDir)

    if os.environ.has_key(ENV_COMMUNICATION):
        g_Communication = os.environ[ENV_COMMUNICATION]
        print "Found %s='%s'" % (ENV_COMMUNICATION, g_Communication)
    else:
        g_Communication = None
    if g_Communication == VALUE_PVM:
        CheckPvmEnvironment()
    elif g_Communication == VALUE_MPI:
        CheckMpiEnvironment()
    else:
        raise t_AppException(p_Message = ("environment variable %s must be either '%s' or '%s'" %
                                          ( ENV_COMMUNICATION, VALUE_MPI, VALUE_PVM )))

    # check PAMINA executable
    if not os.path.exists(g_PaminaExecutable):
        raise t_AppException(p_Message = "cannot find PAMINA executable '%s'" % g_PaminaExecutable)
    
    print "Found PAMINA executable '%s'" % g_PaminaExecutable

    # check route converter executable
    if not os.path.exists(g_RouteConverterExecutable):
        raise t_AppException(p_Message = "cannot find route converter executable '%s'" % g_RouteConverterExecutable)
    
    print "Found route converter executable '%s'" % g_RouteConverterExecutable

    # check planner executable
    if not os.path.exists(g_PlannerExecutable):
        raise t_AppException(p_Message = "cannot find planner executable '%s'" % g_PlannerExecutable)

    print "Found planner executable '%s'" % g_PlannerExecutable

    if not os.path.exists(g_ConfigTemplateFile):
        raise t_AppException(p_Message = "cannot find configuration file template '%s'" % g_ConfigTemplateFile)

    print "Found configuration file template = '%s'" % g_ConfigTemplateFile
                            

# -----------------------------------------------------------------------------------

def DeleteFile(p_Path):

    print "Deleting file '%s'..." % p_Path
    try:
        posix.unlink(p_Path)

    except IOError, v_IOError:
        raise t_AppException(p_Message="while deleting file",
                             p_Filename = p_Path, p_IOError = v_IOError)
    print "File '%s' successfully deleted." % p_Path
    
# -----------------------------------------------------------------------------------

def MoveFile(p_OldPath, p_NewPath):

    print "Moving file '%s' to '%s'..." % (p_OldPath, p_NewPath)
    try:
        posix.rename(p_OldPath, p_NewPath)

    except IOError, v_IOError:
        raise t_AppException(p_Message="while moving file",
                             p_Filename = p_OldPath, p_IOError = v_IOError)
    print "File '%s' successfully moved." % p_OldPath
    
# -----------------------------------------------------------------------------------

def DeleteParallelFeedbackIndexFiles(p_Path):

    v_IndexFile = p_Path + EXTENSION_UNKNOWN_ARCH + EXTENSION_INDEX
    if os.path.exists(v_IndexFile):
        DeleteFile(v_IndexFile)

    v_IndexFile = p_Path + "." + g_Arch + EXTENSION_INDEX
    if os.path.exists(v_IndexFile):
        DeleteFile(v_IndexFile)
        
# -----------------------------------------------------------------------------------

def RenameParallelFeedbackFiles():

    v_ParallelEdgeTimeFile = ReplaceVariables(g_EdgeTimeFileTemplate) + EXTENSION_EDGE_TIME
    v_OldParallelEdgeTimeFile = v_ParallelEdgeTimeFile + EXTENSION_OLD_TIME
    v_NewParallelEdgeTimeFile = v_ParallelEdgeTimeFile + EXTENSION_NEW_TIME

    if os.path.exists(v_NewParallelEdgeTimeFile):
        if os.path.exists(v_ParallelEdgeTimeFile):
            if os.path.exists(v_OldParallelEdgeTimeFile):
                DeleteFile(v_OldParallelEdgeTimeFile)
            MoveFile(v_ParallelEdgeTimeFile, v_OldParallelEdgeTimeFile)
            DeleteParallelFeedbackIndexFiles(v_ParallelEdgeTimeFile)
        MoveFile(v_NewParallelEdgeTimeFile, v_ParallelEdgeTimeFile)

    v_ParallelNodeTimeFile = ReplaceVariables(g_NodeTimeFileTemplate) + EXTENSION_NODE_TIME
    v_OldParallelNodeTimeFile = v_ParallelNodeTimeFile + EXTENSION_OLD_TIME
    v_NewParallelNodeTimeFile = v_ParallelNodeTimeFile + EXTENSION_NEW_TIME

    if os.path.exists(v_NewParallelNodeTimeFile):
        if os.path.exists(v_ParallelNodeTimeFile):
            if os.path.exists(v_OldParallelNodeTimeFile):
                DeleteFile(v_OldParallelNodeTimeFile)
            MoveFile(v_ParallelNodeTimeFile, v_OldParallelNodeTimeFile)
            DeleteParallelFeedbackIndexFiles(v_ParallelNodeTimeFile)
        MoveFile(v_NewParallelNodeTimeFile, v_ParallelNodeTimeFile)
        
# -----------------------------------------------------------------------------------

def GetArchivePath(p_Path):

    v_FileComponents = os.path.split(p_Path)
    v_Filename = v_FileComponents[1] + EXTENSION_GZIP
    return os.path.join(g_ArchiveDir, v_Filename)
    
# -----------------------------------------------------------------------------------

def ArchiveFile(p_Path):

    print "Compress and archive file '%s'..." % p_Path

    v_ArchivePath = GetArchivePath(p_Path)
    v_Command = "gzip -c %s > %s" % (p_Path, v_ArchivePath)
    v_Return = os.system(v_Command)
    if v_Return != 0:
        raise t_AppException(p_Message = "gzip returned exit code %d " % v_Return,
                             p_Filename = p_Path)
    print "File '%s' successfully compressed." % p_Path

    DeleteFile(p_Path)

# -----------------------------------------------------------------------------------


def LoadFileFromArchiveIfNeccessary(p_Path):

    print "Checking file '%s'..." % p_Path

    if os.path.exists(p_Path):
        return

    if not g_CheckArchive:
        raise t_AppException(p_Message = "neccessary file in archive, but option check-archive not active",
                             p_Filename = p_Path)

    v_ArchivePath = GetArchivePath(p_Path)

    print "Checking archived file '%s'..." % v_ArchivePath

    if not os.path.exists(v_ArchivePath):
        raise t_AppException(p_Message = "file not found: neither in work directory nor in archive",
                             p_Filename = p_Path)

    print "Loading file '%s' from archive..." % p_Path
            
    v_Command = "gunzip -c %s > %s" % (v_ArchivePath, p_Path)
    v_Return = os.system(v_Command)
    if v_Return != 0:
        raise t_AppException(p_Message = "gunzip returned exit code %d " % v_Return,
                             p_Filename = v_ArchivePath)
    print "File '%s' successfully uncompressed." % v_ArchivePath

    DeleteFile(v_ArchivePath)


# -----------------------------------------------------------------------------------

def ArchivePlannerFiles():

    v_PreviousTransimsPlanFile = ReplaceVariables(g_TransimsPlanFileTemplate,
                                                  p_Iteration = g_PreviousIteration)

    if os.path.exists(v_PreviousTransimsPlanFile):
        ArchiveFile(v_PreviousTransimsPlanFile)


    v_PreviousPaminaPlanFile = ReplaceVariables(g_PaminaPlanFileTemplate,
                                                p_Iteration = g_PreviousIteration)

    v_TransimsLinkStatFile = v_PreviousPaminaPlanFile + EXTENSION_TRANSIMS_LINK_STAT

    if os.path.exists(v_TransimsLinkStatFile):
        ArchiveFile(v_TransimsLinkStatFile)
    
# -----------------------------------------------------------------------------------

def ArchiveRouteConverterFiles():

    v_TransimsPlanFile = ReplaceVariables(g_TransimsPlanFileTemplate,
                                          p_Iteration = g_Iteration)

    v_TransimsLinkHitStatFile = v_TransimsPlanFile + EXTENSION_TRANSIMS_LINK_HIT_STAT

    if os.path.exists(v_TransimsLinkHitStatFile):
        ArchiveFile(v_TransimsLinkHitStatFile)

    
# -----------------------------------------------------------------------------------

def ArchivePaminaFiles():

    v_PaminaPlanFile = ReplaceVariables(g_PaminaPlanFileTemplate,
                                        p_Iteration = g_Iteration)

    if os.path.exists(v_PaminaPlanFile):
        ArchiveFile(v_PaminaPlanFile)

    v_CompletedRoutesStatFile = v_PaminaPlanFile + EXTENSION_COMPLETED_ROUTES_STAT

    if os.path.exists(v_CompletedRoutesStatFile):
        ArchiveFile(v_CompletedRoutesStatFile)
        
    v_TurnCountStatFile = v_PaminaPlanFile + EXTENSION_TURN_COUNT_STAT

    if os.path.exists(v_TurnCountStatFile):
        ArchiveFile(v_TurnCountStatFile)

    v_RouteByOriginStatFile = v_PaminaPlanFile + EXTENSION_ROUTE_BY_ORIGIN_STAT

    if os.path.exists(v_RouteByOriginStatFile):
        ArchiveFile(v_RouteByOriginStatFile)


# -----------------------------------------------------------------------------------

def ReplaceVariables(p_String, p_Iteration=None):

    if p_Iteration == None:
        p_Iteration = g_Iteration
        
    v_IterationString = g_IterationStringFormat % p_Iteration

    v_IterationPattern = re.compile(r"__ITERATION__")
    v_WorkDirPattern = re.compile(r"%W")

    v_ConfigString = re.sub(v_IterationPattern, v_IterationString, p_String)
    v_ConfigString = re.sub(v_WorkDirPattern, g_WorkingDir, v_ConfigString)

    return v_ConfigString
    
# -----------------------------------------------------------------------------------

def ReadConfig():

    global g_Config, g_IterationStringFormat, g_PaminaHomeDir, g_TransimsMapBaseFilename
    global g_ArchiveDir, g_TransimsPlanFileTemplate, g_PaminaPlanFileTemplate
    global g_EdgeTimeFileTemplate, g_NodeTimeFileTemplate
    global g_ReplanningFraction, g_ReductionFraction, g_UpdateAge
    global g_InitialAge, g_LinearAgeDependant
    
    print "Reading config file template '%s'..." % g_ConfigTemplateFile

    g_Config = t_Configuration(g_ConfigTemplateFile)

    g_IterationStringFormat = g_Config.findInGroup(CONFIG_GROUP_ITERATION,
                                                   CONFIG_ITERATION_STRING_FORMAT,
                                                   DEFAULT_ITERATION_STRING_FORMAT)

    print "Using iteration string format = '%s'" % g_IterationStringFormat
    
    v_PaminaHomeDir = g_Config.findInGroup(CONFIG_GROUP_ITERATION,
                                           CONFIG_PAMINA_HOME_DIR,
                                           DEFAULT_PAMINA_HOME_DIR)

    g_PaminaHomeDir = ReplaceVariables(v_PaminaHomeDir)

    print "Using PAMINA home directory = '%s'" % g_PaminaHomeDir
    
    v_ArchiveDir = g_Config.findInGroup(CONFIG_GROUP_ITERATION,
                                        CONFIG_ARCHIVE_DIR)

    g_ArchiveDir = ReplaceVariables(v_ArchiveDir)

    print "Using archive directory = '%s'" % g_ArchiveDir

    g_TransimsPlanFileTemplate = g_Config.findInGroup(CONFIG_GROUP_ROUTE_CONVERTER,
                                                      CONFIG_TRANSIMS_PLANFILE_TEMPLATE)

    print "Using TRANSIMS plan file template = '%s'" % g_TransimsPlanFileTemplate

    g_PaminaPlanFileTemplate = g_Config.findInGroup(CONFIG_GROUP_ROUTER,
                                                    CONFIG_PAMINA_PLANFILE_TEMPLATE)

    print "Using PAMINA plan file template = '%s'" % g_PaminaPlanFileTemplate

    g_EdgeTimeFileTemplate = g_Config.findInGroup(CONFIG_GROUP_TRAFFICDATA,
                                                  CONFIG_EDGE_TIME_FILE_TEMPLATE)

    print "Using edge feedback file template = '%s'" % g_EdgeTimeFileTemplate

    g_NodeTimeFileTemplate = g_Config.findInGroup(CONFIG_GROUP_TRAFFICDATA,
                                                  CONFIG_NODE_TIME_FILE_TEMPLATE)

    print "Using node feedback file template = '%s'" % g_NodeTimeFileTemplate

    v_TransimsMapBaseFilename = g_Config.findInGroup(CONFIG_GROUP_PLANNER,
                                                     CONFIG_TRANSIMS_MAP_BASE_FILENAME,
                                                     DEFAULT_PAMINA_HOME_DIR)

    g_TransimsMapBaseFilename = ReplaceVariables(v_TransimsMapBaseFilename)
    
    g_ReplanningFraction = float(g_Config.findInGroup(CONFIG_GROUP_PLANNER,
                                                      CONFIG_REPLANNING_FRACTION,
                                                      DEFAULT_REPLANNING_FRACTION))

    g_ReductionFraction = float(g_Config.findInGroup(CONFIG_GROUP_PLANNER,
                                                     CONFIG_REDUCTION_FRACTION,
                                                     DEFAULT_REDUCTION_FRACTION))

    g_UpdateAge = int(g_Config.findInGroup(CONFIG_GROUP_PLANNER,
                                           CONFIG_UPDATE_AGE,
                                           DEFAULT_UPDATE_AGE))

    g_InitialAge = int(g_Config.findInGroup(CONFIG_GROUP_PLANNER,
                                            CONFIG_INITIAL_AGE,
                                            DEFAULT_INITIAL_AGE))

    g_LinearAgeDependant = int(g_Config.findInGroup(CONFIG_GROUP_PLANNER,
                                                    CONFIG_LINEAR_AGE_DEPENDANT,
                                                    DEFAULT_LINEAR_AGE_DEPENDANT))


# -----------------------------------------------------------------------------------

def PrepareConfig():

    global g_ConfigPrepared

    if g_ConfigPrepared:
        return

    v_PaminaHomeDir = ReplaceVariables(g_PaminaHomeDir)
    g_Config.setInGroup(CONFIG_GROUP_ITERATION, CONFIG_PAMINA_HOME_DIR,
                        "\"%s\"" % v_PaminaHomeDir)

#    v_TransimsMapDir = ReplaceVariables(g_TransimsMapDir)
#    g_Config.setInGroup(CONFIG_GROUP_ITERATION, CONFIG_TRANSIMS_MAP_DIR,
#                        "\"%s\"" % v_TransimsMapDir)

    v_PaminaPlanFileTemplate = ReplaceVariables(g_PaminaPlanFileTemplate)
    g_Config.setInGroup(CONFIG_GROUP_ROUTER, CONFIG_PAMINA_PLANFILE_TEMPLATE,
                        "\"%s\"" % v_PaminaPlanFileTemplate)
                        
    v_TransimsPlanFileTemplate = ReplaceVariables(g_TransimsPlanFileTemplate)
    g_Config.setInGroup(CONFIG_GROUP_ROUTE_CONVERTER, CONFIG_TRANSIMS_PLANFILE_TEMPLATE,
                        "\"%s\"" % v_TransimsPlanFileTemplate)

    print "Writing config file '%s'..." % g_ConfigFile

    g_Config.writeToFile(g_ConfigFile, p_Terminator=";")
    
    g_ConfigPrepared = 1
    

# -----------------------------------------------------------------------------------

def ExecutePlanner():

    print "Checking main input files for planner..."

    if g_PreviousIteration > 0:
        v_FeedbackFile = ReplaceVariables(g_PaminaPlanFileTemplate, g_PreviousIteration) + ".tls"
        LoadFileFromArchiveIfNeccessary(v_FeedbackFile)
        v_FeedbackParameter = "-F %s" % v_FeedbackFile
    else:
        v_FeedbackParameter = ""
        
    v_PlanInputFile = ReplaceVariables(g_TransimsPlanFileTemplate, g_PreviousIteration)
    LoadFileFromArchiveIfNeccessary(v_PlanInputFile)

    v_ReplanningFraction = g_ReplanningFraction
    if g_PreviousIteration == 0:
        print "Initial planner OD-Matrix: replanning fraction set to 1.0."
        v_ReplanningFraction = 1.0
    
    v_PlanOutputFile = ReplaceVariables(g_TransimsPlanFileTemplate, g_Iteration)

    v_RandomSeed = int(time.time())

    v_Command = "%s -v %s -f %f -d %f -i %d %s -g %d %s %s -I %s -O %s -r %d -b %d %s" \
                % (g_PlannerExecutable,
                   g_TransimsMapBaseFilename,     # -v
                   v_ReplanningFraction,          # -f
                   g_ReductionFraction,           # -d
                   g_Iteration,                   # -i
                   g_UpdateAge and "-a" or "",
                   g_InitialAge,                  # -g
                   g_LinearAgeDependant and "-l" or "", 
                   v_FeedbackParameter,           # -F
                   v_PlanInputFile,               # -I
                   v_PlanOutputFile,              # -O
                   v_RandomSeed,                  # -r
                   g_BinaryMode,                  # -b
                   g_RedirectSuffix)
    
                               
    print "Executing planner %s..." % v_Command
    
    v_Return = os.system(v_Command)

    if v_Return != 0:
        raise t_AppException(p_Message = "planner returned exit code %d " % v_Return)

    print "Route converter exited successfully."

# -----------------------------------------------------------------------------------

def ExecuteRouteConverter():

    print "Checking main input files for route converter..."
    
    v_PlanInputFile = ReplaceVariables(g_TransimsPlanFileTemplate, g_Iteration)
    LoadFileFromArchiveIfNeccessary(v_PlanInputFile)

    PrepareConfig()

    v_Command = "%s %s" \
                % (g_RouteConverterExecutable,
                   g_RedirectSuffix)

    print "Executing route converter..."
    
    v_Return = os.system(v_Command)

    if v_Return != 0:
        raise t_AppException(p_Message = "route converter returned exit code %d " % v_Return)

    print "Route converter exited successfully."

# -----------------------------------------------------------------------------------

def ExecutePamina():

    print "Checking main input files for PAMINA..."

    v_PaminaPlanFile = ReplaceVariables(g_PaminaPlanFileTemplate)
    LoadFileFromArchiveIfNeccessary(v_PaminaPlanFile)
    
    PrepareConfig()
        
    v_Command = "%s %s" % \
                (g_PaminaExecutable,
                 g_RedirectSuffix)

    print "Executing PAMINA microsimulation..."
    
    v_Return = os.system(v_Command)

    if v_Return != 0:
        raise t_AppException(p_Message = "PAMINA returned exit code %d " % v_Return)

    print "PAMINA exited successfully."


# -----------------------------------------------------------------------------------

def CheckPathAndArchive(p_Path):

    if os.path.exists(p_Path):
        return 1

    v_ArchivePath = GetArchivePath(p_Path)

    if os.path.exists(v_ArchivePath):
        return 2

    return 0

# -----------------------------------------------------------------------------------

def CheckStartPoint():

    global g_SkipPlanner, g_SkipRouteConverter
    
    print "Checking start point for first iteration..."

    # check if we have to run planner: TRANSIMS planfile of current iteration must not exist

    v_TransimsPlanFile = ReplaceVariables(g_TransimsPlanFileTemplate)

    v_Location = CheckPathAndArchive(v_TransimsPlanFile)
    if  v_Location > 0:
        print "Found (%d) TRANSIMS plan file '%s' -> skip planner" % (v_Location, v_TransimsPlanFile)
        g_SkipPlanner = 1

    v_PaminaPlanFile = ReplaceVariables(g_PaminaPlanFileTemplate)

    v_Location = CheckPathAndArchive(v_PaminaPlanFile)
    if  v_Location > 0:
        print "Found (%d) PAMINA plan file '%s' -> skip route converter" % (v_Location, v_PaminaPlanFile)
        g_SkipRouteConverter = 1

    
# -----------------------------------------------------------------------------------

def Main():

    global g_ConfigPrepared, g_Iteration, g_IterationString
    global g_NextIteration, g_PreviousIteration
    global g_StdOut, g_RedirectSuffix

    print "Starting RunIterativeReplanning.pl ..."
    
    ReadConfig()
    
    CheckEnvironment()

    g_Iteration = g_FirstIteration

    while g_Iteration <= g_LastIteration:

        g_ConfigPrepared = 0

        print "Starting iteration %d" % g_Iteration

        if (g_RedirectStdOut):
            g_StdOut = ReplaceVariables(os.path.join(g_WorkingDir, DEFAULT_LOG_FILE))
            print "Standard output of this iteration will be logged %s" % g_StdOut
            g_RedirectSuffix = " >> %s 2>&1" % g_StdOut
            backup(g_StdOut)

        if g_DetectRestartPoint:
            CheckStartPoint()

        g_PreviousIteration = g_Iteration - 1
        g_NextIteration = g_Iteration + 1

        if g_Iteration != g_FirstIteration or not g_SkipPlanner:
            ExecutePlanner()
        ArchivePlannerFiles()

        if g_Iteration != g_FirstIteration or not g_SkipRouteConverter:
            ExecuteRouteConverter()
        ArchiveRouteConverterFiles()

        RenameParallelFeedbackFiles()
        ExecutePamina()
        ArchivePaminaFiles()

        g_Iteration = g_Iteration + 1

    print "Terminating RunIterativeReplanning.pl successfully..."


# -----------------------------------------------------------------------------------

if __name__ == '__main__':
    try:

        v_Options = [ (OPTION_FIRST_ITERATION, "index of first iteration"),
                      (OPTION_LAST_ITERATION, "index of last iteration"),
                      (OPTION_SKIP_PLANNER, "skip planner during first iteration"),
                      (OPTION_SKIP_ROUTE_CONVERTER, "skip route-converter during first iteration"),
                      (OPTION_CONFIG_TEMPLATE, "name of configuration file template"),
                      (OPTION_WORKING_DIRECTORY, "working directory"),
                      (OPTION_REDIRECT_STDOUT, "redirect standard output to a file"),
                      (OPTION_DETECT_RESTART_POINT, "detect the restart point for the first iteration"),
                      (OPTION_CHECK_ARCHIVE, "check archive and unzip files from there if neccessary")]

        v_OptionHandler = t_OptionHandler(v_Options)
        
        v_OptionHandler.parse()

        print "Evaluating command line parameters..."

        g_Iteration = 0
        g_FirstIteration = v_OptionHandler.getInteger(OPTION_FIRST_ITERATION, DEFAULT_FIRST_ITERATION)
        g_LastIteration = v_OptionHandler.getInteger(OPTION_LAST_ITERATION, DEFAULT_LAST_ITERATION)
        g_SkipPlanner = v_OptionHandler.isDefined(OPTION_SKIP_PLANNER)
        g_SkipRouteConverter = v_OptionHandler.isDefined(OPTION_SKIP_ROUTE_CONVERTER)
        g_ConfigTemplateFile = v_OptionHandler.getParameter(OPTION_CONFIG_TEMPLATE, DEFAULT_CONFIG_TEMPLATE)
        g_WorkingDir = v_OptionHandler.getParameter(OPTION_WORKING_DIRECTORY, posix.getcwd())
        g_DetectRestartPoint = v_OptionHandler.isDefined(OPTION_DETECT_RESTART_POINT)
        g_CheckArchive = v_OptionHandler.isDefined(OPTION_CHECK_ARCHIVE)
        g_RedirectStdOut = v_OptionHandler.isDefined(OPTION_REDIRECT_STDOUT)
        
        g_InitialAge = 0
        
        print "Found working directory = '%s'" % g_WorkingDir
        
        Main()
	sys.exit(0)
        

    except IOError:
        printExceptionMessages()
	sys.exit(1)
            
    except t_EnvException:
        printExceptionMessages()
        v_OptionHandler.listOptions();
	sys.exit(2)

    except t_AppException, Exception:
        printExceptionMessages()
	sys.exit(1)

