/*      file     : ParSim/TParSimMaster.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TParSimMaster.C,v 1.20 2005/03/25 11:07:34 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#ifdef QUANTIFY
#  include "quantify.h"
#endif

#include <TrafficDataTypes.h>
#include <TrafficDataConfig.h>
#include <TStream.h>
#include <TEnvironment.h>
#include <TConfiguration.h>
#include <TParallel.h>
#include <TObjectManager.h>
#include <TDynamicBalancer.h>
#include <TRouter.h>
#include <TLocationInfo.h>
#include <TAdjacenceInfo.h>
#include <TTrafficDataManager.h>
#include <TApplicationSlave.h>
#include <TSamplingDescription.h>
#include <TAggregationInfo.h>
#include <TAggregationResult.h>
#include <TStatManager.h>
#include <TMultilaneStat.h>
#include <TLoadStat.h>

#include "ParSimConfig.h"

#include "TParSimMaster.h"

// ***DEFINES************************************************************************* 

/// Default maximum number of timesteps executed in a simulation run.
#define PARSIM_DEFAULT_MAXTIMESTEPS          1000L

/// Default minimum completion (=ratio of executed route plans) which has to executed for the simulation to terminate.
#define PARSIM_DEFAULT_MINCOMPLETION         0.99

#define PARSIM_DEFAULT_SEQUENCELENGTH        50L
#define PARSIM_DEFAULT_GRIDLOCKTIMEOUT       0L

#define PARSIM_DEFAULT_MULTILANESTATDUMP     0L
#define PARSIM_DEFAULT_MULTILANESTATINTERVAL 50L

#define PARSIM_DEFAULT_PERFSTATDUMP          0L
#define PARSIM_DEFAULT_PERFSTATINTERVAL      500L

#define MULTILANE_STAT_EXTENSION             ".mls"
#define PERF_STAT_EXTENSION                  ".perf"

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TApplicationMaster
// =================================================================================== 

TParSimMaster::TParSimMaster(TApplicationSlave *theApplicationSlave) :

ParentClass(theApplicationSlave)

{
    
    SumSites = 0;
    SumObjects = 0;
    SumSpeed = 0;
    MultilaneSamplingDescription = 0;
    MultilaneAggregationInfo = 0;
    MultilaneStatOutput = 0;
    PerfStatOutput = 0;
    MultilaneStatOutput = 0;
    Configuration->SetDefault(TConfigEntry(PARSIM_GROUP, TRAFFIC_NAME_STARTTIME,
                                           TRAFFIC_DEFAULT_STARTTIME));
    Configuration->SetDefault(TConfigEntry(PARSIM_GROUP, PARSIM_NAME_MAXTIMESTEPS,
                                           PARSIM_DEFAULT_MAXTIMESTEPS));
    Configuration->SetDefault(TConfigEntry(PARSIM_GROUP, PARSIM_NAME_GRIDLOCKTIMEOUT,
                                           PARSIM_DEFAULT_GRIDLOCKTIMEOUT));
    Configuration->SetDefault(TConfigEntry(PARSIM_GROUP, 
                                           PARSIM_NAME_MULTILANESTATINTERVAL,
                                           PARSIM_DEFAULT_MULTILANESTATINTERVAL));
    Configuration->SetDefault(TConfigEntry(PARSIM_GROUP, 
                                           PARSIM_NAME_MULTILANESTATDUMP,
                                           PARSIM_DEFAULT_MULTILANESTATDUMP));
    Configuration->SetDefault(TConfigEntry(PARSIM_GROUP, 
                                           PARSIM_NAME_PERFSTATINTERVAL,
                                           PARSIM_DEFAULT_PERFSTATINTERVAL));
    Configuration->SetDefault(TConfigEntry(PARSIM_GROUP, PARSIM_NAME_SEQUENCELENGTH,
                                           PARSIM_DEFAULT_SEQUENCELENGTH));
    Configuration->SetDefault(TConfigEntry(PARSIM_GROUP, PARSIM_NAME_MINCOMPLETION, 
                                           PARSIM_DEFAULT_MINCOMPLETION));
    Configuration->SetDefault(TConfigEntry(PARSIM_GROUP, PARSIM_NAME_PERFSTATDUMP, 
                                           PARSIM_DEFAULT_PERFSTATDUMP));
}

// =================================================================================== 

TParSimMaster::~TParSimMaster(void)

{
    DeleteObject(PerfStatOutput);
    DeleteObject(MultilaneStatOutput);
    DeleteObject(MultilaneSamplingDescription);
    DeleteObject(MultilaneAggregationInfo);
}

// =================================================================================== 

bool
TParSimMaster::ActivateStatistics(void)

{
    
    // start multilane CA statistics...
    MultilaneStatInterval = (int) Configuration->GetLong(PARSIM_GROUP, 
                                                         PARSIM_NAME_MULTILANESTATINTERVAL);
    if (MultilaneStatInterval > 0)
    {
        MultilaneSamplingDescription = 
            new TSamplingDescription(TMultilaneStat::ClassInfo->GetID(), SF_AGGREGATE,
                                     Parallel->GetLocalCPNID(), SAMPLING_ID_GLOBAL_CA_STAT);
        MultilaneAggregationInfo = new TAggregationInfo(this, MultilaneSamplingDescription);
        MultilaneSamplingDescription->SetCollectInterval(MultilaneStatInterval);
        StatManager->RegisterAggregation(MultilaneAggregationInfo);
        StatManager->StartSampling(MultilaneSamplingDescription);

        bool MultilaneDumpFile = Configuration->Exists(PARSIM_GROUP, PARSIM_NAME_MULTILANESTATDUMP);
        if (MultilaneDumpFile) {
            char    Pattern[512];
            TString Filename;
            
            MultilaneStatOutput = new TStream;
            sprintf(Pattern, "%%R%s", MULTILANE_STAT_EXTENSION);
            Filename = Environment->ReplacePathTokens(Pattern);
            if (!MultilaneStatOutput->Open(Filename, "w"))
            {
                Error("ActivateStatistics: cannot open multilane ca stat file %s", (char *) Filename);
                return FALSE;
            }
            
        }
    }   

    // start the performance statistics...
    PerfStatInterval = (int) Configuration->GetLong(PARSIM_GROUP, 
                                                    PARSIM_NAME_PERFSTATINTERVAL);
    if (PerfStatInterval != INVALID AND 
        Configuration->Exists(PARSIM_GROUP, PARSIM_NAME_PERFSTATFILE))
    {
        char    Pattern[512];
        TString Filename;

        PerfStatOutput = new TStream;
        sprintf(Pattern, "%%R%s", PERF_STAT_EXTENSION);
        Filename = Environment->ReplacePathTokens(Pattern);
        if (!PerfStatOutput->Open(Filename, "w"))
        {
            Error("ActivateStatistics: cannot open performance statistics file %s", (char *) Filename);
            return FALSE;
        }
        PerfStatOutput->Write("#1 : simulation time-step\n");
        PerfStatOutput->Write("#2 : application time-step\n");
        PerfStatOutput->Write("#3 : application time\n");
        PerfStatOutput->Write("#4 : vehicles\n");
        PerfStatOutput->Write("#5 : time per time-step\n");
        PerfStatOutput->Write("#6 : RTR\n");
        PerfStatOutput->Write("#7 : MUPS\n");
        PerfStatOutput->Write("#8 : MOSPS\n");
    }
    StatManager->BroadcastSamplingDescriptions();
    return TRUE;
}

// =================================================================================== 

bool
TParSimMaster::Distribute(void)

{
    TNodeList NodeList;
    TEdgeList EdgeList;

    return (TrafficDataManager->CreateSimulationObjects(&NodeList, &EdgeList) AND
            DistributeSimulationObjects(Parallel->GetPhysicalTopology(),
                                        &NodeList, TLocationInfo::ClassInfo,
                                        &EdgeList, TAdjacenceInfo::ClassInfo,
                                        TrafficDataManager->GetDistributionNodeLoadableMask(),
                                        TrafficDataManager->GetDistributionEdgeLoadableMask()));
}

// =================================================================================== 

TString
TParSimMaster::GetClassName(void) const

{
    return "TParSimMaster";
}

// =================================================================================== 

bool
TParSimMaster::HandleAggregationResult(TAggregationInfo *AggregationInfo, 
                                       TTimeStep &TimeStep)

{
    if (AggregationInfo == MultilaneAggregationInfo)
        return HandleMultilaneResult(AggregationInfo, TimeStep);
    return TStatHandler::HandleAggregationResult(AggregationInfo, TimeStep);
}

// =================================================================================== 

bool
TParSimMaster::HandleMultilaneResult(TAggregationInfo *Info, TTimeStep &TimeStep)

{
    TAggregationResult *Result = 0;
    TMultilaneStat     *Stat = 0;
    double              SumTime;

    if ((Result = (TAggregationResult *) Info->GetResult(TimeStep)) == 0)
    {
        Error("HandleMultilaneResult: cannot retrieve aggregation result");
        return FALSE;
    }
    if ((Stat = (TMultilaneStat *) Result->StatObject) == 0)
    {
        Error("HandleMultilaneResult: cannot retrieve stat object");
        return FALSE;
    }   
    Stat->PrintScreenStat(stderr);
    if (MultilaneStatOutput)
        Stat->PrintLineStat(TimeStep, MultilaneStatOutput->GetFile());
    SumSites = Stat->GetNrOfSites();
    SumObjects = Stat->GetNrOfVehicles();
    SumSpeed = Stat->GetSumV();
    if (SumObjects != PreviousVehicleCount OR
        SumSpeed != PreviousVehicleSpeed)
    {
        PreviousVehicleCount = SumObjects;
        PreviousVehicleSpeed = SumSpeed;
        PreviousVehicleTimeStep = TimeStep;
    }
    SumTime = DynamicBalancer->GetLatestGlobalLoadStat()->GetAverageSumTime();
    if (SumObjects > 0 AND SumTime > 0.0)
    {
        if (PerfStatDump)
            Print("Performance: %.3f MUPS %.3f MOSPS RTR %.3f", 
                  (double) SumSites / 1000000.0 / SumTime,
                  (double) SumObjects / 1000000.0 / SumTime,
                  1.0 / SumTime);
        if (PerfStatOutput)
            PerfStatOutput->Write("%ld %ld %f %ld %f %f %f %f\n",
                                  (long) TimeStep,
                                  (long) TimeStep + (long) StartTimeStep,
                                  (double) ((long) TimeStep + (long) StartTimeStep) / 3600.0,
                                  SumObjects,
                                  SumTime,
                                  1.0 / SumTime,
                                  (double) SumSites / 1000000.0 / SumTime,
                                  (double) SumObjects / 1000000.0 / SumTime);
    }
    Info->DisposeResult(TimeStep);
    return TRUE;
}

// =================================================================================== 

bool
TParSimMaster::Run(void)

{
    double Completion;

#ifdef QUANTIFY
    if (quantify_is_running())
    {
        quantify_stop_recording_data();
        Print("stopped collecting QUANTIFY data");
    }
#endif
    PreviousVehicleCount = -1;
    PreviousVehicleSpeed = -1;
    PreviousVehicleTimeStep = 0;
    if (!ParentClass::Run())
        return FALSE;
    if (!Distribute())
        return FALSE;
    if (!ActivateStatistics())
        return FALSE;
    MinCompletion = Configuration->GetDouble(PARSIM_GROUP, PARSIM_NAME_MINCOMPLETION);
    SequenceLength = Configuration->GetLong(PARSIM_GROUP, PARSIM_NAME_SEQUENCELENGTH);
    PerfStatDump = Configuration->GetLong(PARSIM_GROUP, PARSIM_NAME_PERFSTATDUMP);
    MaxTimeSteps = Configuration->GetLong(PARSIM_GROUP, PARSIM_NAME_MAXTIMESTEPS);
    GridLockTimeOut = Configuration->GetLong(PARSIM_GROUP, PARSIM_NAME_GRIDLOCKTIMEOUT);
    StartTimeStep = Configuration->GetLong(PARSIM_GROUP, TRAFFIC_NAME_STARTTIME);

#ifdef QUANTIFY
    if (quantify_is_running())
    {
        quantify_start_recording_data();
        Print("started collecting QUANTIFY data");
    }
#endif

    Print("This is Pamina III");

    while ((long) ApplicationTimeStep < MaxTimeSteps)
    {
        if (!StartApplicationSequence(ApplicationTimeStep, SequenceLength))
            return FALSE;
        Completion = Router->GetCompletion();
        if (Completion >= MinCompletion)
        {
            Print("%.0f%% of all routes completed or failed -> terminating simulation", 
                  100 * Completion);
            break;
        }
        if (GridLockTimeOut > 0 AND
            ApplicationTimeStep - PreviousVehicleTimeStep >= GridLockTimeOut)
        {
            Print("gridlock (for more than %ld time-steps) detected -> terminating simulation",
                  GridLockTimeOut);
            break;
        }
    }
#ifdef QUANTIFY
    if (quantify_is_running())
    {
        quantify_stop_recording_data();
        Print("stopped collecting QUANTIFY data");
    }
#endif
    return TRUE;
}

// =================================================================================== 
#undef ParentClass
