/*      file     : TrafficData/TAccessory.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TAccessory.C,v 1.5 2005/03/25 12:28:51 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <stdio.h>

#include <TBaseStream.h>
#include <TTransferInfo.h>
#include <TClassInfo.h>
#include <TEdge.h>
#include <TSource.h>
#include <TSink.h>

#include "CAConfig.h"
#include "TMultilaneGrid.h"
#include "TAccessory.h"

// ***DEFINES************************************************************************* 

#define SPLIT_SLACK           1
#define MAX_RELATIVE_LOCATION 0.99
#define MIN_RELATIVE_LOCATION 0.01

// ***GLOBAL VARIABLES****************************************************************  

TClassInfo *TAccessory::ClassInfo = 0;

// ***METHODS************************************************************************* 

TAccessoryConstData::TAccessoryConstData(void)

{
    EdgeID = INVALID_ID;
    RelativeLocation = -1.0;
    AbsoluteLocation = -1.0;
}

// =================================================================================== 
#define ParentClass TParallelListObject
// =================================================================================== 

TAccessory::TAccessory(TParallelID theParallelID, TAccessoryConstData *ConstData) :

ParentClass(ClassInfo, theParallelID)

{
    _Init();
    if (ConstData)
        Const = *ConstData;
    Source = new TSource(theParallelID);
    Sink = new TSink(theParallelID);
}

// =================================================================================== 

TAccessory::~TAccessory(void)

{
    Dispose();
}

// =================================================================================== 

bool
TAccessory::Activate(void)

{
    TEdge          *Edge;
    TMultilaneGrid *Grid;
    long            CSite, CLane;
    long            Offset;
    
    if ((Edge = (TEdge *) TEdge::ClassInfo->GetObjectPtr(Const.EdgeID)) == 0)
    {
        Error("Activate: associated edge not found");
        return FALSE;
    }
    if ((Grid = (TMultilaneGrid *) 
         Edge->GetInfo(TMultilaneGrid::ClassInfo->GetInfoID())) == 0)
    {
        Error("Activate: cannot retrieve grid");
        return FALSE;
    }
    Offset = (long) (Const.RelativeLocation * Grid->GetNrOfSites());
    if (!Grid->ComputeOffset(Offset, CA_DEFAULT_V_MAX, 0, CA_DEFAULT_V_MAX))
    {
        Error("Activate: cannot compute offset for grid %s",
              (char *) Grid->GetInstanceName());
        return FALSE;
    }
    for (CLane = 0; CLane < Grid->GetNrOfLanes(); CLane++)
    {
        for (CSite = 0; CSite < CA_DEFAULT_V_MAX; CSite++)
            Grid->SAccess(CLane, Offset + CSite) |= SITE_EMISSION;
    }
    Source->AssignGrid(Grid, Offset);
    Sink->AssignGrid(Grid, Offset);

    //    Print("Activate");
        
    return TRUE;
}

// =================================================================================== 

bool
TAccessory::AddVehicle(TVehicle *Vehicle)

{
    return Source->AddVehicle(Vehicle);
}

/* =================================================================================== */

void
TAccessory::ComputeRelativeLocation(TEdge *Edge, TInfoID AdjacenceInfoID)

{
    double EdgeLength;

    EdgeLength = Edge->GetLength(AdjacenceInfoID);
    Const.RelativeLocation = Const.AbsoluteLocation / EdgeLength;
    if (Const.RelativeLocation < MIN_RELATIVE_LOCATION)
    {
        IfDebug(DM_CONTROL, Warning("relative location will be corrected to %f", 
                                     MIN_RELATIVE_LOCATION));
        Const.RelativeLocation = MIN_RELATIVE_LOCATION;
    }
    if (Const.RelativeLocation > MAX_RELATIVE_LOCATION)
    {
        IfDebug(DM_CONTROL, Warning("relative location will be corrected to %f", 
                                     MAX_RELATIVE_LOCATION));
        Const.RelativeLocation = MAX_RELATIVE_LOCATION;    
    }
    if (Edge->GetToNodeID(AdjacenceInfoID) != Const.DestNodeID)
        Const.RelativeLocation = 1.0 - Const.RelativeLocation;

    //    Print("ComputeRelativeLocation");
}

/* =================================================================================== */

bool
TAccessory::Decode(TBaseStream *aStream, TTransferInfo *TransferInfo)

{
    if (!ParentClass::Decode(aStream, TransferInfo))
        return FALSE;
    switch(TransferInfo->TransferMode)
    {
    case TM_ConstData:  
        if (!(aStream->Decode(Const.EdgeID) AND
              aStream->Decode(Const.DestNodeID) AND
              aStream->Decode(Const.RelativeLocation) AND
              aStream->Decode(Const.AbsoluteLocation)))
            return FALSE;
        
        // The following lines are used for both transfer modes:
        
    case TM_VarData:
        return (Source->Decode(aStream, TransferInfo) AND
                Sink->Decode(aStream, TransferInfo));

    default:
        Warning("Decode: invalid TransferMode %d", TransferInfo->TransferMode); 
    }
    return TRUE;
}

/* =================================================================================== */

void
TAccessory::Dispose(void)

{
    DeleteObject(Source);
    DeleteObject(Sink);
}

/* =================================================================================== */

bool
TAccessory::Encode(TBaseStream *aStream, TTransferInfo *TransferInfo) const

{
    if (!ParentClass::Encode(aStream, TransferInfo))
        return FALSE;
    switch(TransferInfo->TransferMode)
    {
    case TM_ConstData:  
        if (!(aStream->Encode(Const.EdgeID) AND
              aStream->Encode(Const.DestNodeID) AND
              aStream->Encode(Const.RelativeLocation) AND
              aStream->Encode(Const.AbsoluteLocation)))
            return FALSE;
        
        // The following lines are used for both transfer modes:
        
    case TM_VarData:
        return (Source->Encode(aStream, TransferInfo) AND
                Sink->Encode(aStream, TransferInfo));

    default:
        Warning("Encode: invalid TransferMode %d", TransferInfo->TransferMode); 
    }
    return TRUE;
}

// =================================================================================== 

bool
TAccessory::ExecuteSubTimeStep(const TTimeStep &TimeStep)

{
    return (Source->ExecuteSubTimeStep(TimeStep) AND 
           Sink->ExecuteSubTimeStep(TimeStep));
}

// =================================================================================== 

TString
TAccessory::GetClassName(void) const

{
    return "TAccessory";
}

// =================================================================================== 

TString
TAccessory::GetInstanceName(void) const

{
    char Line[512];
    static char Format[512] = {0} ;

    if (Format[0] == 0)
        sprintf(Format, "(%%s) EdgeID:%s DestNodeID:%s offs:%%d AbsLoc:%%.1f RelLoc:%%4.2lf",
                PARALLEL_ID_PRINTF_FORMAT, PARALLEL_ID_PRINTF_FORMAT);
    sprintf(Line, Format, (char *) ParentClass::GetInstanceName(),
            Const.EdgeID, Const.DestNodeID, Offset, 
            (double) Const.AbsoluteLocation, (double) Const.RelativeLocation);
    return Line;
}

// =================================================================================== 

TAccessory *
TAccessory::GetPtr(TParallelID ID)

{
    return (TAccessory *) ClassInfo->GetObjectPtr(ID);
}

// =================================================================================== 

void
TAccessory::_Init(void)

{
    Offset = -1;
}

// =================================================================================== 

bool
TAccessory::IsInRange(const TDistributedRange *Range) const

{
    return Range->IsInRange(Const.RelativeLocation);
}

// =================================================================================== 
#undef ParentClass
