/*      file     : TrafficData/TCAManager.C
        project  : PAMINA III
        author   : Marcus Rickert
        RCS      : $Id: TCAManager.C,v 1.14 2005/03/25 12:28:51 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <TrafficDataTypes.h>
#include <TrafficDataConfig.h>

#include <TConfiguration.h>
#include <TParallel.h>
#include <TDistributedRange.h> // temp
#include <TEdge.h>
#include <TLocationInfo.h>
#include <TEdgeInfo.h>
#include <TGraphEdgeInfo.h>
#include <TSource.h>
#include <TAccessory.h>
#include <TAccessoryHandler.h>
#include <TObjectManager.h>
#include <TMultilaneGridConfig.h>
#include <TMultilaneGrid.h>
#include <TSimpleIntersection.h>
#include <TMultilaneBoundary.h>
#include <TMultilaneStat.h>
#include <TMultilaneDump.h>
#include <TGridStat.h>
#include <TVehicle.h>
#include <TStatManager.h>
#include <TSamplingDescription.h>
#include <TTrafficTimers.h>

#include "TCAManager.h"

// ***DEFINES************************************************************************* 

#define ROUTER_DEFAULT_QUEUEFEEDBACK 0L

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TBaseObject
// ===================================================================================

TCAManager::TCAManager(TEdgeList *theEdges, TEdgeList *theBoundaryEdges, 
                       TEdgeList *theNonBoundaryEdges)

{
    BeforeFirstTimeStep = TRUE;
    Default = new TMultilaneGridConfig;
    Edges = theEdges;
    BoundaryEdges = theBoundaryEdges;
    NonBoundaryEdges = theNonBoundaryEdges;
    Configuration->SetDefault(TConfigEntry(CA_GROUP, CA_NAME_INITDENSITY, 
                                           CA_DEFAULT_INITDENSITY));
    Configuration->SetDefault(TConfigEntry(CA_GROUP, CA_NAME_PASSINGPROHIBITION, 
                                           CA_DEFAULT_PASSINGPROHIBITION));
    Configuration->SetDefault(TConfigEntry(CA_GROUP, CA_NAME_STARTDUMP, 
                                           CA_DEFAULT_STARTDUMP));
    Configuration->SetDefault(TConfigEntry(CA_GROUP, CA_NAME_STOPDUMP, 
                                           CA_DEFAULT_STOPDUMP));
    Configuration->SetDefault(TConfigEntry(CA_GROUP, CA_NAME_DUMPWIDTH, 
                                           CA_DEFAULT_DUMPWIDTH));
    Configuration->SetDefault(TConfigEntry(CA_GROUP, CA_NAME_MINSITES, 
                                           CA_DEFAULT_MINSITES));
    Configuration->SetDefault(TConfigEntry(CA_GROUP, CA_NAME_DECPROB,
                                           CA_DEFAULT_DEC_PROB));
    Configuration->SetDefault(TConfigEntry(CA_GROUP, CA_NAME_SPEEDLIMITS,
                                           CA_DEFAULT_SPEEDLIMITS));
    Configuration->SetDefault(TConfigEntry(ROUTER_GROUP, ROUTER_NAME_QUEUEFEEDBACK,
                                           ROUTER_DEFAULT_QUEUEFEEDBACK));
    debug(Print(DM_LIFESPAN, "CONSTRUCTOR"));    
}

// ===================================================================================

TCAManager::~TCAManager(void)

{
    debug(Print(DM_LIFESPAN, "DESTRUCTOR"));    
    DeleteObject(Default);
}

// ===================================================================================

bool
TCAManager::Activate(void)

{
    debug(Print(DM_LIFESPAN, "Activate"));    
    ObjectManager->RegisterClass(CI_FLAG_TABLE, TMultilaneGrid::ClassInfo,
                                 CLASS_EDGE, CLASS_MULTILANE_GRID, this);
    ObjectManager->RegisterClass(0, TMultilaneBoundary::ClassInfo,
                                 0, CLASS_MULTILANE_BOUNDARY, this);
    ObjectManager->RegisterClass(0, TMultilaneStat::ClassInfo,
                                 0, CLASS_MULTILANE_STAT, this);
    ObjectManager->RegisterClass(0, TGridStat::ClassInfo,
                                 CLASS_EDGE, CLASS_GRID_STAT, this);
    ObjectManager->RegisterClass(0, TMultilaneDump::ClassInfo,
                                 CLASS_EDGE, CLASS_MULTILANE_DUMP, this);
    ObjectManager->RegisterClass(CI_FLAG_TABLE, TVehicle::ClassInfo,
                                 0, CLASS_VEHICLE, this);
    return TRUE;
}

// =================================================================================== 

bool
TCAManager::Collect(TSamplingDescription *Description)

{
    if (Description->GetStatObjectClassID() == TMultilaneStat::ClassInfo->GetID())
    {
        TMultilaneStat *Stat = new TMultilaneStat;

        Stat->SetTimeStep(Description->GetTimeStep());
        if (!CollectMultilaneStat(Stat))
        {
            Error("Collect: cannot collect multilane stat");
            return FALSE;
        }
        Description->GetObjectList()->Append(Stat);
        return TRUE;
    }
    if (Description->GetStatObjectClassID() == TGridStat::ClassInfo->GetID())
    {
        if (!CollectGridStat(Description->GetTimeStep(), Description->GetObjectList()))
        {
            Error("Collect: cannot collect grid stat");
            return FALSE;
        }
        //      Print("collect grid stat for %s", (char *) Description->GetInstanceName());
        return TRUE;
    }
    if (Description->GetStatObjectClassID() == TMultilaneDump::ClassInfo->GetID())
    {
        if (!CollectMultilaneDump(Description->GetTimeStep(), 
                                  Description->GetFrame(), Description->GetObjectList()))
        {
            Error("Collect: cannot collect multilane dump");
            return FALSE;
        }
        return TRUE;
    }
    return TStatHandler::Collect(Description);
}

// =================================================================================== 

bool
TCAManager::CollectGridStat(const TTimeStep &, TStatObjectList *StatList)

{
    TInfoID InfoID;
    TEnumerator< TMultilaneGrid > CGrid(TMultilaneGrid::ClassInfo->GetTable());

    InfoID = TGridStat::ClassInfo->GetInfoID();
    CGrid.Enumerate();
    while (CGrid())
    {
        TGridStat *Stat;
        if ((Stat = (TGridStat *) CGrid->GetInfo(InfoID)) == 0)
        {
            Error("CollectGridStat: cannot retrieve grid stat for %s",
                  (char *) CGrid->GetObjectName());
            return FALSE;
        }
        Stat->SetFraction(((double) CGrid->GetNrOfLocalSites()) / CGrid->GetNrOfSites());
        StatList->Append(Stat);
        CGrid++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TCAManager::CollectMultilaneDump(TTimeStep &TimeStep, const TRectangle &Frame,
                                 TStatObjectList *List)

{
    TEdgeEnum       CEdge;
    TMultilaneGrid *Grid;
    TMultilaneDump *Dump;
    TGraphEdgeInfo *GraphInfo;
    TInfoID         GraphInfoID;
    TInfoID         GridInfoID;

    GraphInfoID = TGraphEdgeInfo::ClassInfo->GetInfoID();
    GridInfoID = TMultilaneGrid::ClassInfo->GetInfoID();
    CEdge.WalksForwardThrough(Edges);
    while (CEdge())
    {
        if ((GraphInfo = (TGraphEdgeInfo *) CEdge->GetInfo(GraphInfoID, TRUE, TRUE)) == 0)
        {
            Error("CollectMultilaneDump: cannot retrieve graph info for %s",
                  (char *) CEdge->GetObjectName());
            return FALSE;
        }
        if (Frame.Contains(GraphInfo->From) OR Frame.Contains(GraphInfo->To))
        {
            double Fraction;

            if ((Grid = (TMultilaneGrid *) CEdge->GetInfo(GridInfoID)) == 0)
            {
                Error("CollectMultilaneDump: %s has no grid", 
                      (char *) CEdge->GetInstanceName());
                return FALSE;
            }
            Dump = new TMultilaneDump(CEdge->GetParallelID());
            Dump->SetTimeStep(TimeStep);
            Fraction = CEdge->GetActiveFraction(Parallel->GetLocalCPNID());
            Dump->SetFraction(Fraction);
            if (!Dump->Collect(Grid))
            {
                Error("CollectMultilaneDump: while collecting from %s",
                      (char *) Grid->GetInstanceName());
                return FALSE;
            }
            IfDebug(DM_STAT_SAMPLE, Print("collect dump from %s", 
                                          (char *) Grid->GetInstanceName()));
            List->Append(Dump);
        }
        CEdge++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TCAManager::CollectMultilaneStat(TMultilaneStat *Stat)

{
    TEdgeEnum       CEdge;
    TMultilaneGrid *Grid;

    CEdge.WalksForwardThrough(Edges);
    while (CEdge())
    {
        if ((Grid = (TMultilaneGrid *) 
             CEdge->GetInfo(TMultilaneGrid::ClassInfo->GetInfoID())) == 0)
        {
            Error("CollectMultilaneStat: %s has no grid", (char *) CEdge->GetInstanceName());
            return FALSE;
        }
        if (!Stat->Collect(Grid))
        {
            Error("CollectMultilaneStat: while collecting from %s",
                  (char *) Grid->GetInstanceName());
            return FALSE;
        }

        IfDebug(DM_STAT_SAMPLE, Print("collect from %s", (char *) Grid->GetInstanceName()));

        CEdge++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TCAManager::ExecuteAccessories(const TTimeStep &TimeStep)

{
    if (TimeStep.GetSubStep() == 0)
    {
        TEnumerator< TAccessory > CAccessory(TAccessory::ClassInfo->GetTable());

        CAccessory.Enumerate();
        while (CAccessory())
        {
            TEdge *Edge = 0;        

            if (TimeEdgesNow(TimeStep))
            {
                if ((Edge = TEdge::GetPtr(CAccessory->GetEdgeID())) == 0)
                {
                    Error("ExecuteAccessories: cannot retrieve edge");
                    return FALSE;
                }
                if (!Edge->StartTimer(EdgeTimerClassInfo->GetInfoID()))
                {
                    Error("ExecuteAccessories: cannot start edge timer");
                    return FALSE;
                }
            }
            if (!CAccessory->ExecuteSubTimeStep(TimeStep))
            {
                Error("ExecuteAccessories: while executing accessory");
                return FALSE;
            }
            if (TimeEdgesNow(TimeStep))
                Edge->StopTimer(EdgeTimerClassInfo->GetInfoID(), 
                                DEFAULT_TIMER_SAMPLE_MASK);
            CAccessory++;
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TCAManager::ExecuteSubTimeStep(TTimeStep &TimeStep, bool BoundariesAvailable)

{
    TEdgeEnum       CEdge;
    TMultilaneGrid *Grid;

    BeforeFirstTimeStep = FALSE;
    
    if (Edges->GetNrOfElements() != 
        BoundaryEdges->GetNrOfElements() + NonBoundaryEdges->GetNrOfElements())
    {
        Error("mismatch of boundary and non-boundary edges");
        return FALSE;
    }
    if (BoundariesAvailable)
    {
        IfDebug(DM_TIME_STEP, Print("ExecuteSubTimeStep : %s for %ld boundary edges",
                                    (char *) TimeStep.GetInstanceName(), 
                                    BoundaryEdges->GetNrOfElements()));
        CEdge.WalksForwardThrough(BoundaryEdges);
        while (CEdge())
        {
            if ((Grid = (TMultilaneGrid *) 
                 CEdge->GetInfo(TMultilaneGrid::ClassInfo->GetInfoID())) == 0)
            {
                Error("ExecuteSubTimeStep: cannot retrieve boundary grid for %s",
                      (char *) CEdge->GetInstanceName());
                return FALSE;
            }
            if (TimeEdgesNow(TimeStep))
                if (!CEdge->StartTimer(EdgeTimerClassInfo->GetInfoID()))
                {
                    Error("ExecuteSubTimeStep: cannot start timer for %s",
                          (char *) CEdge->GetObjectName());
                    return FALSE;
                }
            if (!Grid->ExecuteTimeStep(TimeStep))
            {
                Error("ExecuteSubTimeStep: while executing boundary grid");
                return FALSE;
            }
            if (TimeEdgesNow(TimeStep))
                CEdge->StopTimer(EdgeTimerClassInfo->GetInfoID(), 
                                 DEFAULT_TIMER_SAMPLE_MASK);
            if ((long) TimeStep >= StartDump AND (long) TimeStep < StopDump)
                Grid->AsciiDump(DumpWidth);
            CEdge++;
        }
        
    }
    else
    {
        IfDebug(DM_TIME_STEP, Print("ExecuteSubTimeStep : %s for %ld edges",
                                    (char *) TimeStep.GetInstanceName(),
                                    Edges->GetNrOfElements() - BoundaryEdges->GetNrOfElements()));
        //      CEdge.WalksForwardThrough(Edges);
        CEdge.WalksForwardThrough(NonBoundaryEdges);
        while (CEdge())
        {
            
            //if (!BoundaryEdges->IsInList(CEdge))
            //{
                if ((Grid = (TMultilaneGrid *) 
                     CEdge->GetInfo(TMultilaneGrid::ClassInfo->GetInfoID())) == 0)
                {
                    Error("ExecuteSubTimeStep: cannot retrieve non-boundary grid for %s",
                          (char *) CEdge->GetInstanceName());
                    return FALSE;
                }
                if ((Grid = (TMultilaneGrid *) 
                     CEdge->GetInfo(TMultilaneGrid::ClassInfo->GetInfoID())) == 0)
                {
                    Error("ExecuteSubTimeStep: cannot retrieve grid for %s",
                      (char *) CEdge->GetInstanceName());
                    return FALSE;
                }
                if (TimeEdgesNow(TimeStep))
                    if (!CEdge->StartTimer(EdgeTimerClassInfo->GetInfoID()))
                    {
                        Error("ExecuteSubTimeStep: cannot start timer for %s",
                              (char *) CEdge->GetObjectName());
                        return FALSE;
                    }
                if (!Grid->ExecuteTimeStep(TimeStep))
                {
                    Error("ExecuteSubTimeStep: while executing non-boundary grid");
                    return FALSE;
                }
                if (TimeEdgesNow(TimeStep))
                    CEdge->StopTimer(EdgeTimerClassInfo->GetInfoID(), 
                                     DEFAULT_TIMER_SAMPLE_MASK);
                if ((long) TimeStep >= StartDump AND (long) TimeStep < StopDump)
                    Grid->AsciiDump(DumpWidth);
                //}
            CEdge++;
        }
    }
    return TRUE;
}

// =================================================================================== 

TString
TCAManager::GetClassName(void) const

{
    return "TCAManager";
}

// =================================================================================== 

bool
TCAManager::HandleCreateRequest(TParallelObject *&Object, 
                                const TClassInfo *ClassInfo, 
                                TParallelID ParallelID)

{
    if (ClassInfo == TVehicle::ClassInfo)
        
    {
        if (TVehicle::UnusedVehicles.GetNrOfElements())
        {
            TVehicle *& Vehicle = (TVehicle *&) Object;
            
            Vehicle = TVehicle::UnusedVehicles.GetFirst();
            TVehicle::UnusedVehicles.Delete(Vehicle);
            
            Object->SetParallelID(ParallelID);
        }
        else
            Object = new TVehicle(ParallelID);
    }

    else if (ClassInfo == TMultilaneBoundary::ClassInfo)

        Object = new TMultilaneBoundary(ParallelID);

    else if (ClassInfo == TMultilaneGrid::ClassInfo)
        
        Object = new TMultilaneGrid(ParallelID);

    else if (ClassInfo == TMultilaneStat::ClassInfo)

        Object = new TMultilaneStat(ParallelID);

    else if (ClassInfo == TGridStat::ClassInfo)

        Object = new TGridStat(ParallelID);

    else if (ClassInfo == TMultilaneDump::ClassInfo)

        Object = new TMultilaneDump(ParallelID);

    return TRUE;
}

// =================================================================================== 

bool
TCAManager::HandleLoadRequest(TParallelObject *Object, bool &Loaded)

{
    if (Object->GetClassID() == TMultilaneGrid::ClassInfo->GetID())
    {
        Loaded = TRUE;
        if (!LoadMultilaneGrid((TMultilaneGrid *) Object))
            return FALSE;
    }
    else if (Object->GetClassID() == TVehicle::ClassInfo->GetID())
    {
        Loaded = TRUE;
    }
    else if (Object->GetClassID() == TMultilaneBoundary::ClassInfo->GetID())
    {
        Loaded = TRUE;
    }
    else if (Object->GetClassID() == TMultilaneStat::ClassInfo->GetID())
    {
        Loaded = TRUE;
    }
    else if (Object->GetClassID() == TGridStat::ClassInfo->GetID())
    {
        Loaded = TRUE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TCAManager::InitializeEdgeAfterReception(TEdge *Edge, TDistributedRange *ActiveRange)

{
    TMultilaneGrid    *Grid;
    TAccessoryHandler *AccessoryHandler;

    if ((Grid = (TMultilaneGrid *) Edge->GetInfo(TMultilaneGrid::ClassInfo->GetInfoID(), 
                                                 TRUE, TRUE))) // 08-19-97
    {
        if (!Grid->Activate(*ActiveRange))
        {
            Error("cannot activate %s", (char *) Grid->GetObjectName());
            return FALSE;
        }
        if (InitDensity > 0.0 AND BeforeFirstTimeStep)
            RandomFill(Grid);
        if ((AccessoryHandler = (TAccessoryHandler *) 
             Edge->GetInfo(TAccessoryHandler::ClassInfo->GetInfoID(), 
                           TRUE, TRUE))) // 08-21-97
            if (!AccessoryHandler->Activate(*ActiveRange))
            {
                Error("cannot activate %s", 
                      (char *) AccessoryHandler->GetObjectName());
                return FALSE;
            }
    }
    else
        Warning("%s has no grid", (char *) Edge->GetObjectName());
    return TRUE;
}

// ===================================================================================

bool
TCAManager::LoadMultilaneGrid(TMultilaneGrid *Grid)

{
    TMultilaneGridConfig  Config;
    TEdgeInfo            *EdgeInfo;
    TAdjacenceInfo       *AdjacenceInfo;
    double                Length;
    long                  CASpeedLimit;

    Config = *Default;
    if ((EdgeInfo = (TEdgeInfo *) Grid->GetInfo(TEdgeInfo::ClassInfo->GetInfoID(), 
                                                TRUE, TRUE)) == 0) // 08-19-97
    {
        Error("LoadMultilaneGrid: no edge info available for %s", 
              (char *) Grid->GetObjectName());
        return FALSE;
    }
    if ((AdjacenceInfo = (TAdjacenceInfo *) 
         Grid->GetInfo(TAdjacenceInfo::ClassInfo->GetInfoID())) == 0)
    {
        Error("LoadMultilaneGrid: no adjacence info available for %s", 
              (char *) Grid->GetObjectName());
        return FALSE;
    }
    Config.NrOfLanes = EdgeInfo->NrOfLanes;
    if (SpeedLimits)
    {
        CASpeedLimit = (long) (EdgeInfo->SpeedLimit / SITE_LENGTH + 0.5 + DecProb);
        if (CASpeedLimit == 0)
        {
            Warning("speed limit = %f [m/s] -> Vmax=0 -> increased by one", EdgeInfo->SpeedLimit);
            CASpeedLimit = 1;
        }
        Config.SpeedLimit = min(CASpeedLimit, CA_DEFAULT_V_MAX);
    }

    Length = AdjacenceInfo->GetLength(TLocationInfo::ClassInfo->GetInfoID());

    if (Config.Flags & CA_FLOOR_GRID_LENGTH)
        Config.NrOfSites = max((long) (Length / SITE_LENGTH_IN_CM), 1);
    else
        Config.NrOfSites = (long) (Length / SITE_LENGTH_IN_CM + 0.5);
    if (Config.NrOfSites < MinSites)
    {
        Config.NrOfSites = MinSites;
        //      Warning("number of sites for grid %s increased to %d",
        //              (char *) Grid->GetInstanceName(), MinSites);
    }
    Grid->AssignConfig(&Config);
    //    Print("LoadMultilaneGrid: Sites=%ld", Config.NrOfSites);
    return TRUE;
}

/* =================================================================================== */

bool
TCAManager::RandomFill(TMultilaneGrid *Grid)

{
    long      CLane;
    long      CSite;
    long      SitesToFill;

    Print("RandomFill: %s with density %f", (char *) Grid->GetInstanceName(), InitDensity);
    for (CLane = 0; CLane < Grid->GetNrOfLanes(); CLane++)
    {
        SitesToFill = (long) (InitDensity * (double) Grid->GetNrOfLocalSites() + 0.5);
        for (CSite = 0; CSite < SitesToFill; CSite++)
        {
            TVehicle *NewVehicle = new TVehicle(INVALID);

            NewVehicle->SetMaxVelocityInCAUnits(CA_DEFAULT_V_MAX);
            Grid->SetVehicle(CLane, Grid->GetFirstSite() + CSite, NewVehicle,
                             NewVehicle->GetSitePattern());
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TCAManager::Run(long theStartTimeStep)

{
    StartTimeStep = theStartTimeStep;
    InitDensity = Configuration->GetDouble(CA_GROUP, CA_NAME_INITDENSITY);

    DecProb = Configuration->GetDouble(CA_GROUP, CA_NAME_DECPROB);
    StartDump =  Configuration->GetLong(CA_GROUP, CA_NAME_STARTDUMP);
    StopDump =  Configuration->GetLong(CA_GROUP, CA_NAME_STOPDUMP);
    DumpWidth = Configuration->GetLong(CA_GROUP, CA_NAME_DUMPWIDTH);
    MinSites = Configuration->GetLong(CA_GROUP, CA_NAME_MINSITES);
    SpeedLimits = Configuration->GetLong(CA_GROUP, CA_NAME_SPEEDLIMITS);
    QueueFeedback = Configuration->GetLong(ROUTER_GROUP, ROUTER_NAME_QUEUEFEEDBACK);
    if (Configuration->GetLong(CA_GROUP, CA_NAME_PASSINGPROHIBITION))
        Default->Flags |= CA_PASSING_PROHIBITION;
    if (!(StatManager->RegisterSamplingHandler(TMultilaneStat::ClassInfo->GetID(), this) AND
          StatManager->RegisterSamplingHandler(TGridStat::ClassInfo->GetID(), this) AND
          StatManager->RegisterSamplingHandler(TMultilaneDump::ClassInfo->GetID(), this)))
    {
        Error("Run: cannot register myself as sampling handler");
        return FALSE;
    }
    TSource::StartTimeStep = theStartTimeStep;
    return TRUE;
}

// =================================================================================== 

bool
TCAManager::Sample(TSamplingDescription *Description)

{
    if (Description->GetStatObjectClassID() == TGridStat::ClassInfo->GetID())
    {
        if (!SampleGridStat(Description->GetTimeStep()))
        {
            Error("Collect: cannot collect grid stat");
            return FALSE;
        }
        return TRUE;
    }
    return TStatHandler::Sample(Description);
}

// =================================================================================== 

bool
TCAManager::SampleGridStat(const TTimeStep &TimeStep)

{
    TInfoID InfoID;
    TEnumerator< TMultilaneGrid > CGrid(TMultilaneGrid::ClassInfo->GetTable());

    InfoID = TGridStat::ClassInfo->GetInfoID();
    CGrid.Enumerate();
    while (CGrid())
    {
        TGridStat  *Stat;
        long       AverageDelay = 0;
       
        if ((Stat = (TGridStat *) CGrid->GetInfo(InfoID, TRUE, TRUE)) == 0)
        {
            Error("SampleGridStat: cannot retrieve/create grid stat for %s",
                  (char *) CGrid->GetObjectName());
            return FALSE;
        }
        Stat->SetTimeStep(TimeStep);
        if (QueueFeedback)
        {
            TEdge               *Edge;
            TSimpleIntersection *Intersection = 0;
           
            if ((Edge = TEdge::GetPtr(Stat->GetParallelID())) == 0)
            {
                Error("SampleGridStat: cannot retrieve edge %ld for queue feedback",
                      Stat->GetParallelID());
                return FALSE;
            }
            if ((Intersection = TSimpleIntersection::GetPtr(Edge->GetFromNodeID())))
            {
                TSource *Source;

                if ((Source = Intersection->GetSource(Stat->GetParallelID())))
                    AverageDelay = (long) Source->GetAverageDelay();
            }
        }       
        Stat->Collect(CGrid, AverageDelay);
        CGrid++;
    }
    return TRUE;
}

// =================================================================================== 
#undef ParentClass

