/*      file     : TrafficData/TMultilaneGrid.i.C
        project  : Parallel Toolbox
        author   : Marcus Rickert
        RCS      : $Id: TMultilaneGrid.i.C,v 1.11 2005/03/25 12:28:52 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

#ifndef TMULTILANEGRID_IC
#define TMULTILANEGRID_IC

/* ***INCLUDES************************************************************************ */

#include <TRandom.h>

// ***METHODS*************************************************************************

inline char
TMultilaneGrid::AsciiSite(TSite Site) const

{
    if (SITE_IS_VEHICLE(Site))
    {
        long CV = SITE_GET_VELOCITY(Site);

        if (CV > CA_DEFAULT_V_MAX)
            return '*';
        else
            return '0' + (char) CV;
    }
    else
        if (SITE_IS_OCCUPIED(Site))
            return 'x';
        else
            return '.';
}

// ===================================================================================

inline void
TMultilaneGrid::AssignConfig(const TMultilaneGridConfig *theConfig)

{
    Config = *theConfig;
}

// ===================================================================================

inline TSite &
TMultilaneGrid::FromBoundary(long Lane, long Offset) const

{
    return Sites[Lane][FirstSite + Offset];
}

/* =================================================================================== */

inline int
TMultilaneGrid::GetBndWidth(void) const

{
    return BoundaryWidth;
}

// ===================================================================================

inline int
TMultilaneGrid::GetChangeViewRange(void) const

{
    return Config.ChangeViewRange;
}

// ===================================================================================

inline long
TMultilaneGrid::GetFirstSite(void) const

{
    return FirstSite;
}

// ===================================================================================

inline long
TMultilaneGrid::GetLastSite(void) const

{
    return LastSite;
}

// ===================================================================================

inline int
TMultilaneGrid::GetMoveBndWidth(void) const

{
    return MoveBoundaryWidth;
}

// ===================================================================================

inline int
TMultilaneGrid::GetMoveViewRange(void) const

{
    return Config.MoveViewRange;
}

// ===================================================================================

inline long
TMultilaneGrid::GetNrOfActiveSites(void) const

{
    return LastActiveSite - FirstActiveSite + 1;
}

// ===================================================================================

inline long
TMultilaneGrid::GetNrOfLocalSites(void) const

{
    return LastSite - FirstSite + 1;
}

// ===================================================================================

inline int
TMultilaneGrid::GetNrOfLanes(void) const

{
    return Config.NrOfLanes;
}

// ===================================================================================

inline long
TMultilaneGrid::GetNrOfOccupiedSites(void) const

{
    return Var.OccupiedSites;
}

// ===================================================================================

inline long
TMultilaneGrid::GetNrOfSites(void) const

{
    return Config.NrOfSites;
}

// ===================================================================================

inline bool
TMultilaneGrid::HasMigratingVehicles(const TTimeStep &TimeStep) const

{
    return (LastSite != Config.NrOfSites AND 
            TimeStep.GetSubStep() == 1);
}

// ===================================================================================

inline bool
TMultilaneGrid::IsAsymetric(void) const

{
    return Config.Flags & CA_ASYMMETRIC;
}


// ===================================================================================

inline TSite &
TMultilaneGrid::IncomingBoundary(long Lane, long Offset) const

{
    return SAccess(Lane, IncomingBoundaryOffset + Offset);
}

// ===================================================================================

inline TSite &
TMultilaneGrid::IncomingMigrationArea(long Lane, long Offset) const

{
    return SAccess(Lane, IncomingMigrationAreaOffset + Offset);
}

// ===================================================================================

inline TVehicle * &
TMultilaneGrid::IncomingVehicleMigrationArea(long Lane, long Offset) const

{
    return VAccess(Lane, IncomingMigrationAreaOffset + Offset);
}

// ===================================================================================

inline void
TMultilaneGrid::SetBoundaryBuffer(long CLane, long Offset, TSite Site)

{
    if (BoundaryBufferOffsets[CLane] == INVALID)
    {
        BoundaryBufferOffsets[CLane] = Offset;
        BoundaryBufferSites[CLane] = Site;
    }
    else
        Warning("SetBoundaryBuffer: second vehicle at %ld:%ld", CLane, Site);
}

// ===================================================================================

inline TSite &
TMultilaneGrid::ToBoundary(long Lane, long Offset) const

{
    return Sites[Lane][BoundaryWidth + LastSite + Offset];
}

// ===================================================================================

inline TSite &
TMultilaneGrid::OutgoingBoundary(long Lane, long Offset) const

{
    return SAccess(Lane, OutgoingBoundaryOffset + Offset);
}

// ===================================================================================

inline TSite &
TMultilaneGrid::OutgoingMigrationArea(long Lane, long Offset) const

{
    return SAccess(Lane, OutgoingMigrationAreaOffset + Offset);
}

// ===================================================================================

inline TVehicle * &
TMultilaneGrid::OutgoingVehicleMigrationArea(long Lane, long Offset) const

{
    return VAccess(Lane, OutgoingMigrationAreaOffset + Offset);
}

// ===================================================================================

inline TSite &
TMultilaneGrid::SAccess(long Lane, long Position) const

{
    return _SAccess(Lane, Position);
}

// ===================================================================================

inline TSite &
TMultilaneGrid::SRawAccess(long Lane, long Position) const

{
    return Sites[Lane][Position];
}

// ===================================================================================

inline TVehicle * &
TMultilaneGrid::VAccess(long Lane, long Position) const

{
    return _VAccess(Lane, Position);
}

// ===================================================================================
//                                   Methods for CA update
// ===================================================================================

// Init an empty site with no site data

inline void
TMultilaneGrid::InitSite(long Lane, long Site) const

{
    SAccess(Lane, Site) = 0;
    VAccess(Lane, Site) = 0;
}

// ===================================================================================

inline bool
TMultilaneGrid::RandomDeceleration(TSite) const

{
#if defined(OLDRAND) || defined(OLDRAND2)
#  ifdef CA_MONITOR_RND
    return CountRnd(CLane, Random->IsLower(DecProbLimit));
#  else
    return Random->IsLower(DecProbLimit);
#  endif
#else
#  ifdef CA_MONITOR_RND
    return CountRnd(CLane, Random->IsLower(Config.DecProb));
#  else
    return Random->IsLower(Config.DecProb);
#  endif
#endif

}

// ===================================================================================

// Remove a vehicle from 'Sites' array preserving the site data. Remove the vehicle instance
// from the 'Vehicles' array.

inline void
TMultilaneGrid::RemoveVehicle(long Lane, long Site, bool DisposeVehicle)

{    
    SAccess(Lane, Site) &= SITE_SITE_MASK;
    if (VAccess(Lane, Site))
    {
        Var.OccupiedSites--;
        if (DisposeVehicle)
            delete VAccess(Lane, Site);
        VAccess(Lane, Site) = 0;
    }
    CheckActiveVehicleRangeAfterDeletion();
}

// ===================================================================================

inline void
TMultilaneGrid::RemoveVehicleFromOutgoingMigrationArea(long Lane, long Site)

{    
    RemoveVehicle(Lane, OutgoingMigrationAreaOffset + Site, FALSE);
}

// ===================================================================================

// Move an vehicle with new velocity 'NewV' from arrays 'Sites' and 'Vehicles' to arrays
// 'ToSites' and 'ToVehicles' keeping constant vehicle data, setting the new velocity, and 
// additionally setting 'Other'

inline void
TMultilaneGrid::UpdateSite(long Lane, long From, long NewV) const

{
#if defined(__SGI_CC) || defined (BENCHMARK_GEHM)
    TSite     TempSite;
    TVehicle *TempVehicle;

    TempSite = (_SAccess(Lane, From + NewV) & SITE_SITE_MASK) |
        (_SAccess(Lane, From) & SITE_CONST_VEHICLE_MASK) | SITE_SHIFT_V(NewV);
    _SAccess(Lane, From + NewV) = TempSite;
    if (NewV)
    {
        TempVehicle = _VAccess(Lane, From);
        _VAccess(Lane, From + NewV) = TempVehicle;
        TempSite = _SAccess(Lane, From) & SITE_SITE_MASK;
        _SAccess(Lane, From) = TempSite;
        _VAccess(Lane, From) = 0;
    }
    if (!SITE_IS_VEHICLE(_SAccess(Lane, From + NewV)))
        Error("vehicle disappeared from Sites");
    if (_VAccess(Lane, From + NewV) == 0)
        Error("vehicle disappeared from Vehicles");
#else
    SAccess(Lane, From + NewV) = (SAccess(Lane, From + NewV) & SITE_SITE_MASK) |
        (SAccess(Lane, From) & SITE_CONST_VEHICLE_MASK) |
            SITE_SHIFT_V(NewV);
    if (NewV)
    {
        VAccess(Lane, From + NewV) = VAccess(Lane, From);
        SAccess(Lane, From) = SAccess(Lane, From) & SITE_SITE_MASK;
        VAccess(Lane, From) = 0;
    } 
#endif
}

// ===================================================================================

inline void
TMultilaneGrid::CheckLRVMax(long Lane, long Site) const

{
    TSite *CSite;

    CSite = &SAccess(Lane, Site);
    if (Lane >= SITE_GET_BND_LANE(*CSite))
        *CSite = (*CSite & ~SITE_V_MAX_MASK) |
            SITE_SHIFT_V_MAX(VAccess(Lane, Site)->GetMaxCAVelocity());
}

// ===================================================================================

inline void
TMultilaneGrid::CheckRLVMax(long Lane, long Site) const

{
    TSite *CSite;

    CSite = &SAccess(Lane, Site);
    if (Lane <= SITE_GET_BND_LANE(*CSite))
        *CSite = (*CSite & ~SITE_V_MAX_MASK) |
            SITE_SHIFT_V_MAX(VAccess(Lane, Site)->GetMaxCAVelocity());
}

// ===================================================================================

inline void
TMultilaneGrid::CheckActiveVehicleRangeAfterDeletion(void)

{
    /*
    if (SkipWhenEmpty() AND Var.OccupiedSites == 0)
    {
        FirstActiveSite = FirstSite;
        LastActiveSite = FirstSite;
    }
    */
}

// ===================================================================================

inline void
TMultilaneGrid::CheckActiveVehicleRangeAfterInsertion(long /* Offset */)

{
    /*
    if (SkipWhenEmpty())
    {
        if (FirstActiveSite == LastActiveSite)
        {
            FirstActiveSite = Offset;
            LastActiveSite = Offset + 1;
        }
        else
        {
            if (Offset < FirstActiveSite)
                FirstActiveSite = Offset;
            if (Offset > LastActiveSite - 1)
                LastActiveSite = Offset + 1;
        }
    }
    */
}

// ===================================================================================

// Copy the vehicle dependant data and vehicle from the arrays 'Sites' and 'Vehicles' to
// the arrays 'ToSites' and 'ToVehicles' while preserving the site dependant data of the
// source arrays

inline void
TMultilaneGrid::CombineSites(long Lane, long Site,
                             long ToLane, long ToSite) const

{
    SAccess(ToLane, ToSite) = (SAccess(ToLane, ToSite) & SITE_SITE_MASK) |
        (SAccess(Lane, Site) & SITE_VEHICLE_MASK);
    SAccess(Lane, Site) &= SITE_SITE_MASK;
    VAccess(ToLane, ToSite) = VAccess(Lane, Site);
    VAccess(Lane, Site) = 0;
}

// ===================================================================================

// Reset the site in the destination arrays using the site dependant data in the source arrays

inline void
TMultilaneGrid::MoveEmptySite(long Lane, long Site,
                              long ToLane, long ToSite) const
    
{
    
    SAccess(ToLane, ToSite) = SAccess(Lane, Site) & SITE_SITE_MASK;
    VAccess(ToLane, ToSite) = 0;
}

// ===================================================================================

// Copy the non-temporary vehicle and site dependant data and vehicle from the arrays 
// 'Sites' and 'Vehicles' to the arrays 'ToSites' and 'ToVehicles' overwriting all data 
// in the destination arrays

inline void
TMultilaneGrid::MoveSite(long Lane, long Site,
                         long ToLane, long ToSite) const

{
    SAccess(ToLane, ToSite) = SAccess(Lane, Site) &
        SITE_DATA_MASK & ~SITE_PREV_TS_LANE_CHANGE;
    VAccess(ToLane, ToSite) = VAccess(Lane, Site);
}

// ===================================================================================

inline bool
TMultilaneGrid::SetVehicle(long Lane, long Offset, TVehicle *Vehicle, TSite SitePattern)
    
{
    if (SITE_IS_VEHICLE(SAccess(Lane, Offset)))
    {
        Error("Site offset %ld lane %ld is occupied!", Offset, Lane);
        return FALSE;
    }
    SAccess(Lane, Offset) = SITE_GET_SITE(SAccess(Lane, Offset)) | SitePattern;
    VAccess(Lane, Offset) = Vehicle;
    Var.OccupiedSites++;
    CheckActiveVehicleRangeAfterInsertion(Offset);
    return TRUE;
}

// ===================================================================================

inline bool
TMultilaneGrid::SetMigratedVehicle(long Lane, long Offset, TVehicle *Vehicle, 
                                   TSite SitePattern)
    
{
    if (!SetVehicle(Lane, IncomingMigrationAreaOffset + Offset, Vehicle, SitePattern))
        return FALSE;
    CheckActiveVehicleRangeAfterInsertion(IncomingMigrationAreaOffset + Offset);
    return TRUE;
}

// ===================================================================================

inline bool
TMultilaneGrid::SkipWhenEmpty(void) const

{
    return (Config.Flags & CA_SKIP_WHEN_EMPTY) > 0;
}

// ===================================================================================

inline bool
TMultilaneGrid::VehicleRangeIsActive(long Offset, long Width) const

{
    if (SkipWhenEmpty())
        return !(Offset + Width < FirstActiveSite OR Offset >= LastActiveSite);
    else
        return TRUE;
            
}

// ===================================================================================
#endif
