/*      file     : common/TABTreeElement.C
        project  : General Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TABTreeElement.C,v 1.5 2005/03/25 11:04:59 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <stdlib.h>
#include <stdio.h>

#include <TDebug.h>

#include "TABTreeElement.h"

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#  include "TABTreeElement.i.C"
#endif

// =================================================================================== 
#define ParentClass
// =================================================================================== 

TABTreeElement::TABTreeElement(TABTreeElement *FatherPar, bool IsLeafPar)

{
    Father = FatherPar;
    Leaf = IsLeafPar;
    LeftMostItem = 0;
    ElementsUsed = 0;
    Prev = 0;
    Next = 0;
    debug(Print(DM_AB_TREE, "CONSTRUCTOR IsLeaf=%d", Leaf));
}

// ===================================================================================

TABTreeElement::~TABTreeElement(void)

{
    int i;
    
    if (!IsLeaf())
        for (i = 0; i < ElementsUsed; i++)
            delete References[i];
}


// ===================================================================================

bool
TABTreeElement::Apply(TApplyFunction *Function, void *Context)

{
    int i;

    for (i = 0; i < GetElementsUsed(); i++)
        if (IsLeaf())
        {
            if (!Function(References[i], Context))
                return FALSE;
        }
        else
            if (!GetSon(i)->Apply(Function, Context))
                return FALSE;
    return TRUE;
    
}
        
// ===================================================================================

bool
TABTreeElement::Check(TABTreeElement *TrueFather)

{
    int i;
    
    if (Father != TrueFather)
    {
        fprintf(stderr, "TABTreeElement::Check: wrong father!\n");
        return FALSE;
    }
    if (!IsLeaf())
    {
        if (GetLeftMostItem() != GetSon(0)->GetLeftMostItem())
        {
            fprintf(stderr, "TABTreeElement::Check: wrong LeftMostItem!\n");
            return FALSE;
        }
        for (i = 0; i < GetElementsUsed(); i++)
            if (!GetSon(i)->Check(this))
                return FALSE;
    }
    return TRUE;
}
    
// ===================================================================================

void
TABTreeElement::Delete(int DeletionPoint)

{
    delete References[DeletionPoint];
    Remove(DeletionPoint);
    IfDebug(DM_AB_TREE, Print("son %i deleted", DeletionPoint));
}

// ===================================================================================

bool
TABTreeElement::DeletePossible(void)

{
    return ElementsUsed > TAB_A;
}

// ===================================================================================

void
TABTreeElement::DeleteObjects(void)

{
    int i;

    if (IsLeaf())
        for (i = 0; i < ElementsUsed; i++)
            delete References[i];
    else
        for (i = 0; i < ElementsUsed; i++)
            References[i]->DeleteObjects();
}

// ===================================================================================

TString 
TABTreeElement::GetClassName(void) const

{
    return "TABTreeElement";
}

// ===================================================================================

TBaseObject *
TABTreeElement::GetCompareItem(int Son)
    
{
    if (IsLeaf())
        return References[Son];
    else
        return References[Son]->LeftMostItem;
}

// ===================================================================================

TString 
TABTreeElement::GetInstanceName(void) const

{
    char line[128];

    sprintf(line, "IsLeaf:%d, #ElementsUsed:%d Father:%s ", Leaf, ElementsUsed,
            Father? "non-NULL":"NULL");
    return line;
}

// ===================================================================================

TABTreeElement *
TABTreeElement::GetLeftNeighbour(void)

{
    int SonNumber = 0;

    if (Father AND (SonNumber = Father->GetSonNumber(this)) > 0)
        return Father->GetSon(SonNumber - 1);
    else
        return 0;
}

// ===================================================================================

TABTreeElement *
TABTreeElement::GetRightNeighbour(void)

{
    int SonNumber = 0;

    if (Father AND (SonNumber = Father->GetSonNumber(this)) < Father->GetElementsUsed()-1 )
        return Father->GetSon(SonNumber + 1);
    else
        return 0;
}

// ===================================================================================

int
TABTreeElement::GetSonNumber(TABTreeElement *Son)

{
    int i;
    
    for (i = 0; i < ElementsUsed; i++)
        if (References[i] == Son)
            return i;
    return -1;
}

// ===================================================================================

void
TABTreeElement::Insert(int InsertionPoint, TBaseObject *Data)

{
    int Dest;

#ifdef DCHECK
    if (ElementsUsed == TAB_B)
    {
        Error("Overflow");
        exit (1);
    }
#endif
    for (Dest = ElementsUsed++; Dest > InsertionPoint; Dest --)
        References[Dest] = References[Dest-1];
    References[InsertionPoint] = (TABTreeElement *) Data;
    if (!IsLeaf())
        References[InsertionPoint]->SetFather(this);
    if (InsertionPoint == 0)
        RetrieveLeftmostItem();
    IfDebug(DM_AB_TREE, Print("Insert: InsertionPoint: %d", InsertionPoint));
}

// ===================================================================================

void
TABTreeElement::Merge(TABTreeElement *RightNeighbour)

{
    int i;

    IfDebug(DM_AB_TREE, Print("Merging %d + %d.", ElementsUsed, RightNeighbour->GetElementsUsed()));
    for (i = ElementsUsed; i < ElementsUsed+RightNeighbour->GetElementsUsed(); i++)
    {
        References[i] = RightNeighbour->GetSon(i - ElementsUsed);
        
        if (!IsLeaf())  // MR 1-17-94 discovered by PURIFY
            
            References[i]->SetFather(this);
    }    
    ElementsUsed += RightNeighbour->GetElementsUsed();
    if (IsLeaf())
    {
        Next = RightNeighbour->Next;
        if (Next)
            Next->Prev = this;
    }
    RightNeighbour->Reset();
}

// ===================================================================================

void
TABTreeElement::Remove(int SonNr)

{
    int i;
    
    for (i = SonNr; i < ElementsUsed-1; i++)
        References[i] = References[i+1];
    if (SonNr == 0)
        RetrieveLeftmostItem();
    ElementsUsed--;
}

// ===================================================================================

void
TABTreeElement::Reset(void)

{
    ElementsUsed = 0;
    Prev = 0;
    Next = 0;
}

// ===================================================================================

void
TABTreeElement::ResetLeaf(void)

{
    Leaf = FALSE;
}

// ===================================================================================

bool
TABTreeElement::SetLeftMostItem(TABTreeElement *Son)

{
    if (References[0] == Son)
    {
        LeftMostItem = Son->GetLeftMostItem();
        IfDebug(DM_AB_TREE, Print("*SetLeftMostItem*"));
        return TRUE;
    }
    else
        return FALSE;
}

// ===================================================================================

TABTreeElement *
TABTreeElement::StealLeftmostSon(void)

{
    TABTreeElement *LeftmostSon;

    LeftmostSon = References[0];
    Remove(0);
    TraceLeftmostItem();
    return LeftmostSon;
}

// ===================================================================================

TABTreeElement *
TABTreeElement::StealRightmostSon(void)

{
    TABTreeElement *RightmostSon;

    RightmostSon = References[ElementsUsed-1];
    Remove(ElementsUsed-1);
    return RightmostSon;
}

// ===================================================================================

void
TABTreeElement::RetrieveLeftmostItem(void)

{
    if (IsLeaf())
        LeftMostItem = References[0];
    else
        LeftMostItem = References[0]->LeftMostItem;
}

// ===================================================================================

void
TABTreeElement::Split(TABTreeElement *NewElement)

{
    int i;

    for (i = 0; i<TAB_SPLIT; i++)
        NewElement->Insert(i, References[i+TAB_SPLIT]);
    NewElement->RetrieveLeftmostItem();
    if (IsLeaf())
    {
        NewElement->Next = Next;
        NewElement->Prev = this;
        if (Next)
            Next->Prev = NewElement;
        Next = NewElement;
    }
    ElementsUsed = TAB_SPLIT;
}

// ===================================================================================

void
TABTreeElement::TraceLeftmostItem(void)

{
    TABTreeElement *Son;
    TABTreeElement *Element = this;
    
    do
    {
        Son = Element;
        Element = Element->GetFather();
    }
    while  (Element AND Element->SetLeftMostItem(Son));
}

// =================================================================================== 
#undef ParentClass
