/*      file     : common/TOrthogonalBisection.C
        project  : General Toolbox
        author   : Marcus Rickert
        objects  : TOrthogonalBisection
        RCS      : $Id: TOrthogonalBisection.C,v 1.16 2005/03/25 11:04:59 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <memory.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#ifdef TOOLBOX
#  include <TDebug.h>
#  include <TString.h>
#else
#  define DM_CONTROL 0
#  define DM_FILE    0
#endif

#include "TOrthogonalBisection.h"

// ***DEFINES************************************************************************* 

#define compare(v1, v2) (((v1) < (v2))? -1 : (((v1) > (v2))? 1:0))

#define LengthInKilometer(length) ((length) / 100000.0)

#define MAX_DOUBLE        (1.0E+37)
#define MAX_CLUSTER_HEAP 32

#define CPNDistributionFormatString "ORB_%s_%04d_%04d"

// ***LOCAL VARIABLES***************************************************************** 

long    *SortIndex = 0;
TOBNode *NodeArray = 0;

// ***FUNCTIONS*********************************************************************** 

void
InitNode(TOBNode &aNode)

{
    aNode.Weight = 0.0;
    aNode.X = aNode.Y = 0.0;
    aNode.Flags = 0;
    aNode.ClusterID = NIL;
    aNode.NextNodeInCluster = NIL;
    aNode.EdgeListBase = NIL;
}

// =================================================================================== 

void
InitEdge(TOBEdge &anEdge)

{
    anEdge.Nodes[0] = anEdge.Nodes[1] = NIL;
    anEdge.Weight = 0.0;
    anEdge.Multiplicity = 1;
    anEdge.Flags = 0;
}

// =================================================================================== 

void
InitCPN(TOBCPN &aCPN)

{
    aCPN.FirstCluster = NIL;
    aCPN.NrOfNodes = 0;
    aCPN.NrOfClusters = 0;
    aCPN.NrOfEdges = 0;
}

// =================================================================================== 

int
CompareLogicalEdgeID(const TOBEdge *Edge1, const TOBEdge *Edge2)

{
    return compare(Edge1->ID, Edge2->ID);
}

// =================================================================================== 

int
CompareLogicalNodeReferences(const TOBEdge *Edge1, const TOBEdge *Edge2)

{
    if (Edge1->Nodes[0] == Edge2->Nodes[0])
        return compare(Edge1->Nodes[1], Edge2->Nodes[1]);
    else
        return compare(Edge1->Nodes[0], Edge2->Nodes[0]);
}

// =================================================================================== 

int
CompareLogicalNodeID(const TOBNode *Node1, const TOBNode *Node2)

{
    return compare(Node1->ID, Node2->ID);
}

// =================================================================================== 

int
CompareNodeX(const long *NodeIndex1, const long *NodeIndex2)

{
    return compare(NodeArray[*NodeIndex1].X,
                   NodeArray[*NodeIndex2].X);
}

// =================================================================================== 

int
CompareNodeY(const long *NodeIndex1, const long *NodeIndex2)

{
    return compare(NodeArray[*NodeIndex1].Y,
                   NodeArray[*NodeIndex2].Y);
}

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#include "TOrthogonalBisection.i.C"
#endif

// =================================================================================== 

TOrthogonalBisection::TOrthogonalBisection(void)

{
    NrOfNodes = 0;
    NrOfEdges = 0;
    NrOfClusters = 0;
    NrOfCPN = 0;
    NrOfIncidences = 0;
    Nodes = 0;
    Edges = 0;
    CPN = 0;
    Incidences = 0;
}

// =================================================================================== 

TOrthogonalBisection::~TOrthogonalBisection(void)

{
    if (Nodes)
        delete [] Nodes;
    if (Edges)
        delete [] Edges;
    if (CPN)
        delete [] CPN;
    if (Incidences)
        delete [] Incidences;
}

// =================================================================================== 

bool
TOrthogonalBisection::Allocate(int theMaxNrOfNodes, int theMaxNrOfEdges, 
                               int theMaxNrOfCPN)

{
    int i;

    if (theMaxNrOfNodes <= 0 OR \
        theMaxNrOfEdges <= 0 OR 
        theMaxNrOfCPN <= 0)
    {
        Error("Allocate: cannot bisect for #nodes=%ld #edges=%ld #CPN=%ld",
              theMaxNrOfNodes, theMaxNrOfEdges, theMaxNrOfCPN);
        return FALSE;
    }
    MaxNrOfNodes  = theMaxNrOfNodes;
    MaxNrOfEdges  = theMaxNrOfEdges;
    MaxNrOfCPN = theMaxNrOfCPN;

    if ((Nodes = new TOBNode[MaxNrOfNodes]) == 0 OR
        (Edges = new TOBEdge[MaxNrOfEdges]) == 0 OR
        (Incidences = new TOBIncidence[2 * MaxNrOfEdges]) == 0 OR
        (CPN = new TOBCPN[MaxNrOfCPN]) == 0)
    {
        Error("Allocate: out of heap");
        return FALSE;
    }
    for (i = 0; i < MaxNrOfNodes; i++)
        InitNode(Nodes[i]);
    for (i = 0; i < MaxNrOfEdges; i++)
        InitEdge(Edges[i]);
    for (i = 0; i < MaxNrOfCPN; i++)
        InitCPN(CPN[i]);
    return TRUE;
}

// =================================================================================== 

bool
TOrthogonalBisection::Bisect(int FirstCluster, int LastCluster, 
                             int FirstCPN, int LastCPN)

{
    double MinX, MinY, MaxX, MaxY;
    double LeftWeight, SumWeight;
    double LeftPerformance, RightPerformance;
    double mWeight, Imbalance, ShiftedImbalance;
    long   CNode;
    int    CCluster;
    int    CCPN;
    int    mCPN;
    int    mCluster;
    bool   SplitHorizontally;
    
    MinX = MinY = MAX_DOUBLE;
    MaxX = MaxY = -MAX_DOUBLE;
    SumWeight = 0.0;
    for (CCluster = FirstCluster; CCluster <= LastCluster; CCluster++)
    {
        if (SortedNodes(CCluster).X < MinX)
            MinX = SortedNodes(CCluster).X;
        if (SortedNodes(CCluster).X > MaxX)
            MaxX = SortedNodes(CCluster).X;
        if (SortedNodes(CCluster).Y < MinY)
            MinY = SortedNodes(CCluster).Y;
        if (SortedNodes(CCluster).Y > MaxY)
            MaxY = SortedNodes(CCluster).Y;
        SumWeight += SortedNodes(CCluster).Weight;
    }
    if (LastCPN == FirstCPN)
    {
        // the recursion is terminated if only one CPN is left: all remaining
        // nodes are assigned to that CPN

        CPN[FirstCPN].Weight = SumWeight;
        CPN[FirstCPN].NrOfClusters = LastCluster - FirstCluster + 1;
        CPN[FirstCPN].FirstCluster = NIL;
        for (CCluster = FirstCluster; CCluster <= LastCluster; CCluster++)
        {
            // Insert cluster in cluster list of CPN
            SortedNodes(CCluster).NextClusterOnCPN = (int) CPN[FirstCPN].FirstCluster;
            CPN[FirstCPN].FirstCluster = SortIndex[CCluster];
            // Mark nodes as part of CPN FirstCPN
            SortedNodes(CCluster).CPNID = FirstCPN;
            CNode = SortedNodes(CCluster).NextNodeInCluster;
            while (CNode != NIL)
            {
                Nodes[CNode].CPNID = FirstCPN;
                CNode = Nodes[CNode].NextNodeInCluster;
            }       
        }
    }
    else
    {
        // there is more than one CPN left: prepare recursion
        if (FirstCluster >= LastCluster)
        {
            Error("Bisect: cannot distribute a single node among %ld CPN", 
                  LastCPN - FirstCPN + 1);
            return FALSE;
        }
        // find out wether the node rectangle is heigher than wide:
        SplitHorizontally = MaxY - MinY > MaxX - MinX;

        // sort nodes according to their X or Y coordinates
        qsort(SortIndex + FirstCluster, LastCluster - FirstCluster + 1, 
              sizeof(long), (int(*)(const void*, const void*))
              (SplitHorizontally? CompareNodeY:CompareNodeX));
        if ((LastCPN - FirstCPN + 1) % 2 == 0)
        {
            // the number of CPN is even -> simple split
            mCPN = (FirstCPN + LastCPN) / 2;
        }
        else
        {
            // the number of CPN is odd 
            mCPN = (FirstCPN + LastCPN) / 2;

            // chose between two split potential points:
            // !!!!!STILL MISSING!!!!!
        }
        LeftPerformance = 0.0;
        RightPerformance = 0.0;
        for (CCPN = FirstCPN; CCPN <= mCPN; CCPN++)
            LeftPerformance += CPN[CCPN].Performance;
        for (CCPN = mCPN + 1; CCPN <= LastCPN; CCPN++)
            RightPerformance += CPN[CCPN].Performance;
        LeftWeight = SumWeight * LeftPerformance / (LeftPerformance + RightPerformance);
        mCluster = FirstCluster;
        mWeight = 0.0;
        while (mCluster < LastCluster AND mWeight < LeftWeight)
        {
            mWeight += SortedNodes(mCluster).Weight;
            mCluster++;
        }
        mCluster--;
        if (BisectionFlags & BF_BEST_OF_TWO AND mCluster > FirstCluster)
        {
            Imbalance = fabs(mWeight - LeftWeight);
            ShiftedImbalance = fabs(mWeight - SortedNodes(mCluster - 1).Weight - LeftWeight);
            if (ShiftedImbalance < Imbalance)
            {
                mCluster--;
                NrOfShiftImprovements++;
            }
        }
        if (!(Bisect(FirstCluster, mCluster, FirstCPN, mCPN) AND
              Bisect(mCluster + 1, LastCluster, mCPN + 1, LastCPN)))
            return FALSE;
    }
    return TRUE;    
}

// =================================================================================== 

bool
TOrthogonalBisection::Bisection(int theCurrentNrOfCPN, int theBisectionFlags)

{
    int CCluster;

    BisectionFlags = theBisectionFlags;
    NrOfShiftImprovements = 0;
    CurrentNrOfCPN = theCurrentNrOfCPN;
    if ((SortIndex = new long[NrOfNodes]) == 0)
    {
        Error("Bisect: out of heap");
        return FALSE;
    }
    Print(DM_CONTROL, "bisection of %d clusters (%d nodes) among %d CPN...", 
          NrOfClusters, NrOfNodes, CurrentNrOfCPN);
    for (CCluster = 0; CCluster < NrOfClusters; CCluster++)
        SortIndex[CCluster] = Nodes[CCluster].Cluster;
    NodeArray = Nodes;
    if (!Bisect(0, NrOfClusters-1, 0, CurrentNrOfCPN - 1))
        return FALSE;
    delete [] SortIndex;
    SortIndex = 0;
    return ComputeStatistics();
}

// =================================================================================== 

bool
TOrthogonalBisection::CheckMap(void)

{
    int i;
    long NodeIndex1 = 0, NodeIndex2 = 0;

    Print(DM_CONTROL, "sorting nodes...");
    qsort(Nodes, NrOfNodes, sizeof(TOBNode),
          (int(*)(const void*, const void*))CompareLogicalNodeID);
    Print(DM_CONTROL, "checking node references and computing node weights...");
    for (i = 0; i < NrOfEdges; i++)
    {
        
        if ((NodeIndex1 = GetNodeIndexByID(Edges[i].Nodes[0])) == NIL OR
            (NodeIndex2 = GetNodeIndexByID(Edges[i].Nodes[1])) == NIL)
        {
            Error("edge %ld has invalid node references (%ld, %ld)", Edges[i].ID,
                  Edges[i].Nodes[0], Edges[i].Nodes[1]);
            return FALSE;
        }
        Nodes[NodeIndex1].Flags |= TOB_NODE_FLAG_CONNECTED;
        Nodes[NodeIndex1].Weight += 0.5 * Edges[i].Weight;
        Nodes[NodeIndex2].Flags |= TOB_NODE_FLAG_CONNECTED;
        Nodes[NodeIndex2].Weight += 0.5 * Edges[i].Weight;
    }
    Print(DM_CONTROL, "checking for isolated nodes...");
    for (i = 0; i < NrOfNodes; i++)
        if ((Nodes[i].Flags & TOB_NODE_FLAG_CONNECTED) == 0)
            Warning("node %ld is isolated", Nodes[i].ID);
    return TRUE;
}

// =================================================================================== 

bool
TOrthogonalBisection::Cluster(void)

{
    long CNode, HeapNode;
    long Heap[MAX_CLUSTER_HEAP]; // it is actually used as a stack
    int TopOfHeap;
    long CEdge, CIncidence;
    long ClusterHead;
    int MaxNodesInCluster;
    int CCluster;
    int NodesInCluster;
    long RemoteNode;
    double MaxClusterWeight;
    double SumX, SumY, SumWeight, OverallSumWeight;

    Print(DM_CONTROL, "building incidence lists...");
    for (CEdge = 0; CEdge < NrOfEdges; CEdge++)
    {
        // convert logical node references to physical references

        Edges[CEdge].Nodes[0] = GetNodeIndexByID(Edges[CEdge].Nodes[0]);
        Edges[CEdge].Nodes[1] = GetNodeIndexByID(Edges[CEdge].Nodes[1]);

        InsertIncidentEdge(CEdge, Edges[CEdge].Nodes[0]);
        InsertIncidentEdge(CEdge, Edges[CEdge].Nodes[1]);
    }
    Print(DM_CONTROL, "clustering nodes...");
    CNode = 0;
    while (CNode < NrOfNodes)
    {
        // look for a node that has not been clustered yet
        if ((Nodes[CNode].Flags & TOB_NODE_FLAG_SCANNED) == 0)
        {
            // store the current node as first node in the current cluster
            ClusterHead = CNode;
            Nodes[ClusterHead].Flags |= TOB_NODE_FLAG_SCANNED;
            Nodes[NrOfClusters].Cluster = ClusterHead;
            NrOfClusters++;
            // store the current node in the heap
            TopOfHeap = 0;
            Heap[TopOfHeap++] = ClusterHead;
            while (TopOfHeap)
            {
                // remove top element of heap and make it part of cluster

//              Print("current node %d", CNode);

                HeapNode = Heap[--TopOfHeap];
                if (HeapNode != ClusterHead)
                {
                    // add current node to cluster node list
                    Nodes[HeapNode].NextNodeInCluster = Nodes[ClusterHead].NextNodeInCluster;
                    Nodes[ClusterHead].NextNodeInCluster = HeapNode;
                }
                // go through all incident edges of the current node
                CIncidence = Nodes[HeapNode].EdgeListBase;
                while (CIncidence != NIL)
                {
                    CEdge = Incidences[CIncidence].EdgeID;

//                  Print("current edge %d", CEdge);

                    // check wether edge cannot be split
                    if (Edges[CEdge].Flags & TOB_EDGE_FLAG_NO_SPLIT)
                    {
                        // determine other node (!= CNode) of edge
                        RemoteNode = (Edges[CEdge].Nodes[0] == HeapNode)? 
                            Edges[CEdge].Nodes[1] : Edges[CEdge].Nodes[0];
                        if ((Nodes[RemoteNode].Flags & TOB_NODE_FLAG_SCANNED) == 0)
                        {
                            // add the remote node to the heap
                            Heap[TopOfHeap++] = RemoteNode;
                            Nodes[RemoteNode].Flags |= TOB_NODE_FLAG_SCANNED;
//                          Print("insert node %d into heap", RemoteNode);
                        }
                    }
                    CIncidence = Incidences[CIncidence].Next;
                } // all incident edges
            } // heap not empty
        }
        CNode++;
    }
    OverallSumWeight = 0.0;
    MaxNodesInCluster = 0;
    MaxClusterWeight = 0.0;
    Print(DM_CONTROL, "compressing clusters into single nodes...");
    CCluster = 0;
    while (CCluster < NrOfClusters)
    {
        NodesInCluster = 0;
        SumX = SumY = SumWeight = 0.0;
        CNode = ClusterHead = Nodes[CCluster].Cluster;
        while (CNode != NIL)
        {
            SumX += Nodes[CNode].NodeX;
            SumY += Nodes[CNode].NodeY;
            SumWeight += Nodes[CNode].Weight;
            NodesInCluster++;       
            CNode = Nodes[CNode].NextNodeInCluster;
        }
        OverallSumWeight += SumWeight;
        if (NodesInCluster > 1)
        {
            Nodes[ClusterHead].X = SumX / (double) NodesInCluster;
            Nodes[ClusterHead].Y = SumY / (double) NodesInCluster;
            Nodes[ClusterHead].Weight = SumWeight;
        }
        if (SumWeight > MaxClusterWeight)
            MaxClusterWeight = SumWeight;
        if (NodesInCluster > MaxNodesInCluster)
            MaxNodesInCluster = NodesInCluster;
        CCluster++;
    }
    Print(DM_CONTROL, "there are %d clusters; maximum cluster size is %d node(s) "
          "and %4.1f times the average weight", NrOfClusters, MaxNodesInCluster,
          MaxClusterWeight / (OverallSumWeight / NrOfNodes));
    return TRUE;
}

// =================================================================================== 

bool
TOrthogonalBisection::CompressMap(void)

{
    int Source, Dest;
    int NonSplitableEdges;

    Print(DM_CONTROL, "sorting edges according to first node reference...");
    qsort(Edges, NrOfEdges, sizeof(TOBEdge),
          (int(*)(const void*, const void*)) CompareLogicalNodeReferences);

    NonSplitableEdges = 0;
    Source = 1; 
    Dest = 0;
    Print(DM_CONTROL, "compressing parallel edges...");
    while (Source < NrOfEdges)
    {
        if (CompareLogicalNodeReferences(Edges + Dest, Edges + Source) == 0)
        {
            Edges[Dest].Weight += Edges[Source].Weight;
            if (Edges[Dest].Flags & TOB_EDGE_FLAG_NO_SPLIT AND
                Edges[Dest].Multiplicity == 1)
                NonSplitableEdges++;
            Edges[Dest].Multiplicity++;
            if (Edges[Source].Flags & TOB_EDGE_FLAG_NO_SPLIT)
                Edges[Dest].Flags |= TOB_EDGE_FLAG_NO_SPLIT;
        }
        else
        {
            Dest++;
            if (Source != Dest)
                memcpy(Edges + Dest, Edges + Source, sizeof(TOBEdge));
        }
        Source++;
    }
    NrOfEdges = Dest + 1;
    Print(DM_CONTROL, "new number of edges: %d (%d of which must not be split)", 
          NrOfEdges, NonSplitableEdges);
    return TRUE;
}

// =================================================================================== 

bool
TOrthogonalBisection::ComputeStatistics(void)

{
    int CCPN;
    int CEdge;
    int CPN1, CPN2;
    double SumPerformance, SumWeight;
    double WeightPerPerformance;

    Print(DM_CONTROL, "computing statistics...");
    MinPerformance = MinWeight = MinWeightPerPerformance = MAX_DOUBLE;
    MaxPerformance = MaxWeight = MaxWeightPerPerformance = -MAX_DOUBLE;
    MinNrOfNodesPerCPN = NrOfNodes;
    MaxNrOfNodesPerCPN = 0;
    SumWeight = SumPerformance = 0.0;
    for (CCPN = 0; CCPN < CurrentNrOfCPN; CCPN++)
    {
        CPN[CCPN].NrOfEdges = 0;
        CPN[CCPN].NrOfSplitEdges = 0;
        SumWeight += CPN[CCPN].Weight;
        SumPerformance += CPN[CCPN].Performance;
        if (CPN[CCPN].NrOfNodes < MinNrOfNodesPerCPN)
            MinNrOfNodesPerCPN = CPN[CCPN].NrOfNodes;
        if (CPN[CCPN].NrOfNodes > MaxNrOfNodesPerCPN)
            MaxNrOfNodesPerCPN = CPN[CCPN].NrOfNodes;
        if (CPN[CCPN].Weight < MinWeight)
            MinWeight = CPN[CCPN].Weight;
        if (CPN[CCPN].Weight > MaxWeight)
            MaxWeight = CPN[CCPN].Weight;
        if (CPN[CCPN].Performance < MinPerformance)
            MinPerformance = CPN[CCPN].Performance;
        if (CPN[CCPN].Performance > MaxPerformance)
            MaxPerformance = CPN[CCPN].Performance;
        WeightPerPerformance = CPN[CCPN].Weight / CPN[CCPN].Performance;
        if (WeightPerPerformance < MinWeightPerPerformance)
            MinWeightPerPerformance = WeightPerPerformance;
        if (WeightPerPerformance > MaxWeightPerPerformance)
            MaxWeightPerPerformance = WeightPerPerformance;
    }
    AverageWeightPerPerformance = SumWeight / SumPerformance;
    NrOfSplitEdges = 0;
    for (CEdge = 0; CEdge < NrOfEdges; CEdge++)
    {
        CPN1 = Nodes[Edges[CEdge].Nodes[0]].CPNID;
        CPN2 = Nodes[Edges[CEdge].Nodes[1]].CPNID;
        if (CPN1 != CPN2)
        {
            NrOfSplitEdges += Edges[CEdge].Multiplicity;
            CPN[CPN1].NrOfSplitEdges += Edges[CEdge].Multiplicity;
            CPN[CPN2].NrOfSplitEdges += Edges[CEdge].Multiplicity;
        }
        else
        {
            CPN[CPN1].NrOfEdges += Edges[CEdge].Multiplicity;
        }
    }
    MaxEfficiency = AverageWeightPerPerformance / MaxWeightPerPerformance;
    MaxSpeedup = CurrentNrOfCPN * MaxEfficiency;
    return TRUE;
}

// =================================================================================== 
#ifndef TOOLBOX
void
TOrthogonalBisection::Error(const char *Format, ...) const

{
    va_list argptr;
    char    line[1024];

    va_start(argptr, Format);
    vsprintf(line, Format, argptr);
    fprintf(stderr, "ERROR: %s\n", line);
    va_end(argptr);
}
#endif
// =================================================================================== 
#ifdef TOOLBOX
TString
TOrthogonalBisection::GetClassName(void) const

{
    return "TOrthogobalBisection";
}
#endif
// =================================================================================== 

int
TOrthogonalBisection::GetCPNIndexByNodeID(long NodeID) const

{
    long NodeIndex;

    if ((NodeIndex = GetNodeIndexByID(NodeID)) == NIL)
    {
        Error("GetCPNIndexByNodeID: node ID %ld not found", NodeID);
        return NIL;
    }
    return Nodes[NodeIndex].CPNID;
}

// =================================================================================== 

long
TOrthogonalBisection::GetNodeIndexByID(long theID) const

{
    TOBNode SearchNode;
    TOBNode *NodeFound;

    SearchNode.ID = theID;
    if ((NodeFound = (TOBNode *) bsearch(&SearchNode, Nodes, NrOfNodes, sizeof(TOBNode), 
                 (int(*)(const void*, const void*)) CompareLogicalNodeID)))
        return NodeFound - Nodes;
    else
        return NIL;
}

// =================================================================================== 

void
TOrthogonalBisection::InsertIncidentEdge(long PEdgeID, long PNodeID)

{
    Incidences[NrOfIncidences].EdgeID = PEdgeID;
    Incidences[NrOfIncidences].Next = Nodes[PNodeID].EdgeListBase;
    Nodes[PNodeID].EdgeListBase = NrOfIncidences;
    NrOfIncidences++;
}

// =================================================================================== 

bool
TOrthogonalBisection::Prepare(void)

{
    if (!(CheckMap() AND
          CompressMap() AND
          Cluster()))
        return FALSE;


    return TRUE;
}

// ===================================================================================
#ifndef TOOLBOX
void
TOrthogonalBisection::Print(const char * /* Format */ , ...)

{
    va_list argptr;

    va_start(argptr, Format);
    va_end(argptr);
}
#endif
// ===================================================================================
#ifndef TOOLBOX
void
TOrthogonalBisection::Print(unsigned long, const char * /* Format */ , ...)

{
    va_list argptr;

    va_start(argptr, Format);
    va_end(argptr);
}
#endif
// =================================================================================== 

bool
TOrthogonalBisection::PutCPN(double thePerformance, int Count)

{
    int j;

    for (j = 0; j < Count; j++)
    {
        if (NrOfCPN >= MaxNrOfCPN)
        {
            Error("too many CPN");
            return FALSE;
        }
        CPN[NrOfCPN].Performance = thePerformance;
        NrOfCPN++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TOrthogonalBisection::PutEdge(long theID, long Node1, long Node2, 
                              double theWeight, bool NoSplit)

{
    if (NrOfEdges >= MaxNrOfEdges)
    {
        Error("too many edges");
        return FALSE;
    }
    Edges[NrOfEdges].ID = theID;
    Edges[NrOfEdges].Weight = theWeight;
    if (NoSplit)
        Edges[NrOfEdges].Flags |= TOB_EDGE_FLAG_NO_SPLIT;
    if (Node1 < Node2)
    {
        Edges[NrOfEdges].Nodes[0] = Node1;
        Edges[NrOfEdges].Nodes[1] = Node2;
    }
    else
    {
        Edges[NrOfEdges].Nodes[0] = Node2;
        Edges[NrOfEdges].Nodes[1] = Node1;
    }
    //        Print("PutEdge with weight %f", theWeight);
    NrOfEdges++;
    return TRUE;
}

// =================================================================================== 

bool
TOrthogonalBisection::PutNode(long theID, double theX, double theY, double theWeight)

{
    if (NrOfNodes >= MaxNrOfNodes)
    {
        Error("too many nodes");
        return FALSE;
    }
    Nodes[NrOfNodes].ID = theID;
    Nodes[NrOfNodes].Weight = theWeight;
    Nodes[NrOfNodes].X = Nodes[NrOfNodes].NodeX = theX;
    Nodes[NrOfNodes].Y = Nodes[NrOfNodes].NodeY = theY;
    NrOfNodes++;
    //    Print("PutNode with weight %f", theWeight);
    return TRUE;
}

// =================================================================================== 
#ifndef TOOLBOX
void
TOrthogonalBisection::Warning(const char *Format, ...) const

{
    va_list argptr;
    char    line[1024];

    va_start(argptr, Format);
    vsprintf(line, Format, argptr);
    fprintf(stderr, "Warning: %s\n", line);
    va_end(argptr);
}
#endif
// =================================================================================== 

bool
TOrthogonalBisection::WriteClustersAndNonClusters(const char *BaseFilename)

{
    long CCluster;
    long CNode;
    char Filename[512];
    FILE *ClusterFile = 0;
    FILE *NonClusterFile = 0;

    Print(DM_FILE, "writing cluster and non-cluster files files...");
    sprintf(Filename, "%s_clusters", BaseFilename);
    if ((ClusterFile = fopen(Filename, "w")) == 0)
    {
        Error("cannot open '%s'");
        return FALSE;
    }
    sprintf(Filename, "%s_non-clusters", BaseFilename);
    if ((NonClusterFile = fopen(Filename, "w")) == 0)
    {
        Error("cannot open '%s'");
        fclose(ClusterFile);
        return FALSE;
    }
    for (CCluster = 0; CCluster < NrOfClusters; CCluster++)
    {
        CNode = Nodes[CCluster].Cluster;
        if (Nodes[CNode].NextNodeInCluster == NIL)
            fprintf(NonClusterFile, "%f %f\n", 
                    LengthInKilometer(Nodes[CNode].NodeX),
                    LengthInKilometer(Nodes[CNode].NodeY));
        else
        {
            while (CNode != NIL)
            {
                fprintf(ClusterFile, "%f %f\n", 
                        LengthInKilometer(Nodes[CNode].NodeX),
                        LengthInKilometer(Nodes[CNode].NodeY));
                CNode = Nodes[CNode].NextNodeInCluster;
            }
        }
    }
    fclose(ClusterFile);
    fclose(NonClusterFile);
    return TRUE;
}

// =================================================================================== 

bool
TOrthogonalBisection::WriteCPNDistributionFiles(const char *BaseFilename)

{
    int   CCPN;
    long  CNode, CCluster;
    FILE *Ascii = 0;
    char  Filename[512];

    Print(DM_FILE, "writing CPN distribution files...");
    for (CCPN = 0; CCPN < CurrentNrOfCPN; CCPN++)
    {
        sprintf(Filename, CPNDistributionFormatString, BaseFilename, 
                CCPN, CurrentNrOfCPN);
        if ((Ascii = fopen(Filename, "w")) == 0)
        {
            Error("WriteCPNDistributionFiles: cannot open '%s' for output", Filename);
            return FALSE;
        }
        CCluster = CPN[CCPN].FirstCluster;
        while (CCluster != NIL)
        {
            fprintf(Ascii, "%ld %f %f\n", Nodes[CCluster].ID, 
                    LengthInKilometer(Nodes[CCluster].NodeX),
                    LengthInKilometer(Nodes[CCluster].NodeY));
            CNode = Nodes[CCluster].NextNodeInCluster;
            while (CNode != NIL)
            {
                fprintf(Ascii, "%ld %f %f\n", Nodes[CNode].ID, 
                        LengthInKilometer(Nodes[CNode].NodeX), 
                        LengthInKilometer(Nodes[CNode].NodeY));
                CNode = Nodes[CNode].NextNodeInCluster;
            }
            CCluster = Nodes[CCluster].NextClusterOnCPN;
        }
        fclose(Ascii);
    }
    return TRUE;
}

// =================================================================================== 

void
TOrthogonalBisection::WriteStatHeader(FILE *Ascii) const

{
    int Col = 1;

    fprintf(Ascii, "# column description\n");
    fprintf(Ascii, "# col %2d : number of CPN\n", Col++);
    fprintf(Ascii, "# col %2d : min performance\n", Col++);
    fprintf(Ascii, "# col %2d : max performance\n", Col++);
    fprintf(Ascii, "# col %2d : min weight\n", Col++);
    fprintf(Ascii, "# col %2d : max weight\n", Col++);
    fprintf(Ascii, "# col %2d : min number of nodes\n", Col++);
    fprintf(Ascii, "# col %2d : max number of nodes\n", Col++);
    fprintf(Ascii, "# col %2d : min weight/performance\n", Col++);
    fprintf(Ascii, "# col %2d : max weight/performance\n", Col++);
    fprintf(Ascii, "# col %2d : number of split edges\n", Col++);
    fprintf(Ascii, "# col %2d : max efficiency\n", Col++);
    fprintf(Ascii, "# col %2d : max speed-up\n", Col++);
}

// =================================================================================== 

void
TOrthogonalBisection::WriteStatLine(FILE *Ascii) const

{
    fprintf(Ascii, "%d %f %f %f %f %d %d %f %f %d %f %f\n",
            CurrentNrOfCPN,
            MinPerformance,
            MaxPerformance,
            MinWeight,
            MaxWeight,
            MinNrOfNodesPerCPN,
            MaxNrOfNodesPerCPN,
            MinWeightPerPerformance,
            MaxWeightPerPerformance,
            NrOfSplitEdges,
            MaxEfficiency,
            MaxSpeedup);            
}

// =================================================================================== 
