/*      file     : parallel/TParallel.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TParallel.C,v 1.26 2005/03/25 11:06:20 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <stdlib.h>

#include <ParallelConfig.h>

#include <TString.h>
#include <TEnvironment.h>
#include <TConfiguration.h>

#include <TCPNInfo.h>
#include <TTopology.h>
#include <THostInfo.h>
#include <TCommunication.h>
#include <TTrivialParallelControl.h>
#include <TBaseMessage.h>
#ifdef USE_PVM
#  include <TPVMParallelControl.h>
#else
#  ifdef USE_MPI
#    include <TMPIParallelControl.h>
#  endif
#endif

#include <TParallel.h>

// ***DEFINES************************************************************************* 

#define PARALLEL_PREFIX "Parallel."

#define P_EVENT_CONFIGURATION     1

#define PARALLEL_BROADCAST_CONFIG ((long) FALSE)

// ***GLOBAL VARIABLES****************************************************************  

TParallel *Parallel = 0;

// ***METHODS************************************************************************* 

#ifdef NO_INLINE
#  include "TParallel.i.C"
#endif

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TParallel::TParallel(void)

{
    _Init();
    if (Environment == 0)
    {
        Fail("no TEnvironment active");
    }
    if (Parallel)
    {
        Fail("another instance of TParallel already active");
    }
    Parallel = this;
    Configuration->Set(TConfigEntry(PARALLEL_GROUP, PARALLEL_NAME_BROADCAST_CONFIG, 
                                    PARALLEL_BROADCAST_CONFIG));
#ifdef USE_PVM
    Control = new TPVMParallelControl;
#elif USE_MPI
    Control = new TMPIParallelControl(Environment->pargc(), Environment->pargv());
#else
    Control = new TTrivialParallelControl;
#endif
    debug(Print(DM_LIFESPAN, "CONSTRUCTOR"));
}

// =================================================================================== 

TParallel::~TParallel(void)

{
    debug(Print(DM_LIFESPAN, "DESTRUCTOR"));
    Deactivate();
}

// =================================================================================== 

bool
TParallel::Activate(void)

{
    debug(Print(DM_LIFESPAN, "Activate"));
    BroadcastConfigActive = 
        Configuration->GetLong(PARALLEL_GROUP, PARALLEL_NAME_BROADCAST_CONFIG);
    if (Communication == 0)
    {
        Error("no TCommunication active");
        return FALSE;
    }
    if (!Communication->RegisterRecipient(MTAG_PARALLEL, this))
    {
        Error("cannot register myself as message recipient");
        return FALSE;
    }
    if (Environment == 0)
    {
        Error("no TEnvironment active");
        return FALSE;
    }
    if (!Control->Activate())
    {
        Error("Activate: cannot activate low-level driver");
        return FALSE;
    }
    if (GetMaxNrOfCPN() == -1)
    {
        Error("Activate: cannot determine maximum number of CPN");
        return FALSE;
    }
    return TRUE;
}

// ===================================================================================

bool
TParallel::AddAllCPN(void)

{
    const THostInfoList *HostInfos;
    THostInfoEnum        CHost;

    if (Control->GetCurrentNrOfCPN() < Control->GetMaxNrOfCPN())
    {
        if ((HostInfos = Control->GetHostInfoList()) == 0)
        {
            Error("AddAllCPN: cannot retrieve host info list");
            return FALSE;
        }       
        CHost.WalksForwardThrough(HostInfos);
        while (CHost())
        {
            if (!AddMaxCPN(CHost))
                return FALSE;
            CHost++;
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TParallel::AddCPN(THostInfo *aHostInfo, int NrOfInstances)

{
    if (!Control->AddCPN(aHostInfo, NrOfInstances))
        return FALSE;
    if (NrOfInstances > 0)
    {
        if (!Control->SendChangesInTopology())
            return FALSE;
    }
    return TRUE;
}

// ===================================================================================

bool
TParallel::AddMaxCPN(THostInfo *aHostInfo)

{
    return AddCPN(aHostInfo, aHostInfo->MaxNrOfCPN - aHostInfo->NrOfCPN);
}

// =================================================================================== 

bool
TParallel::BroadcastConfiguration(void)

{
    TBaseMessage *Message = 0;

    if (!Communication->GetNewMessage(Message, DEST_ALL_BUT_MYSELF,
                                      MTAG_PARALLEL, MTYPE_PARALLEL_CONTROL))
    {
        Error("BroadcastConfiguration: cannot get message");
        return FALSE;
    }
    debug(Print(DM_CONFIG, "broadcasting configuration"));
    return (Message->Event(P_EVENT_CONFIGURATION) AND
            Environment->GetConfiguration()->Encode(*Message) AND
            Communication->Commit(Message));
}

// =================================================================================== 

bool
TParallel::Deactivate(void)

{
    if (Control)
    {
        Control->Disenroll();
        DeleteObject(Control);
    }
    Parallel = 0;
    return TRUE;
}

// =================================================================================== 

TString
TParallel::GetClassName(void) const

{
    return "TParallel";
}

// =================================================================================== 

const TCPNInfoList *
TParallel::GetCPNInfoList(void) const

{
    return Control->GetCPNInfoList();
}

// =================================================================================== 

TString
TParallel::GetInstanceName(void) const

{
    if (Control)
        return Control->GetInstanceName();
    else
        return "?";
}

// =================================================================================== 

int
TParallel::GetMaxNrOfCPN(void) const

{
    if (ForcedNrOfCPN == INVALID)
        return Control->GetMaxNrOfCPN();
    else
        return ForcedNrOfCPN;
}

// =================================================================================== 

bool
TParallel::HandleEvent(TEventID theEventID, TBaseMessage *theMessage, int &Handled)

{
    switch (theEventID)
    {
    case P_EVENT_CONFIGURATION:
        return ReceiveConfiguration(*theMessage, Handled);

    default:
        Error("event ID %ld not handled in message %s",
              theEventID, (char *) theMessage->GetInstanceName());
        break;
    }
    return TRUE;
}

// =================================================================================== 

void
TParallel::_Init(void)

{
    ForcedNrOfCPN = INVALID;
    Control = 0;
    Network = 0;
    debug(Print(DM_LIFESPAN, "Hello World"));
}

// =================================================================================== 

bool
TParallel::ReceiveConfiguration(TBaseMessage &aMessage, int &Handled)

{
    const TCPNInfo *LocalInfo;

    if (!Environment->GetConfiguration()->Decode(aMessage))
    {
        Error("cannot receive configuration");
        return FALSE;
    }
    Environment->InterpretConfig();
    Print(/* DM_CONFIG,*/ "received configuration");
    if ((LocalInfo = GetLocalCPNInfo()))
        Environment->SetProcessName(LocalInfo->Name);
    Handled = TRUE;
    return TRUE;
}

// =================================================================================== 

bool
TParallel::ShowCPN(void) const

{
    const TCPNInfoList *CPN;
    TCPNInfoEnum        CCPN;
    int                 i;

    if ((CPN = Control->GetCPNInfoList()) == 0)
    {
        Error("ShowHosts: cannot retrieve host info list");
        return FALSE;
    }
    Print("CPN Configuration:");
    i = 0;
    CCPN.WalksForwardThrough(CPN);
    while (CCPN())
    {
        Print("CPN %3d: %s %s", i + 1, (char *) CCPN->GetInstanceName(),
              (CCPN->ID == GetLocalCPNID())? "<- myself":"");
        i++;
        CCPN++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TParallel::ShowHosts(void) const

{
    const THostInfoList *HostInfos;
    THostInfoEnum        CHost;
    int                  i;

    if ((HostInfos = Control->GetHostInfoList()) == 0)
    {
        Error("ShowHosts: cannot retrieve host info list");
        return FALSE;
    }
    Print("Host Configuration:");
    i = 0;
    CHost.WalksForwardThrough(HostInfos);
    while (CHost())
    {
        Print("Host %2d: %s", i + 1, (char *) CHost->GetInstanceName());
        i++;
        CHost++;
    }
    return TRUE;
}

// ===================================================================================

bool
TParallel::Start(void)

{
    if (IsMaster())
    {
        if (BroadcastConfigActive)
            if (!BroadcastConfiguration())
                return FALSE;
    }
    return Control->Sync();
}

// =================================================================================== 
#undef ParentClass

