/*       file     : parallel/TTopology.h
         project  : Parallel Toolbox 2.x
         author   : Marcus Rickert
        RCS      : $Id: TTopology.h,v 1.8 2005/03/25 11:06:21 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

#ifndef TTOPOLOGY_H
#define TTOPOLOGY_H

// ***INCLUDES************************************************************************ 

#include <stdlib.h>

#include <TBaseObject.h>
#include <TArray.h>
#include <TDLList.h>

#include <ParallelTypes.h>

#include "TTopologyUser.h"

// ***FORWARDS************************************************************************

class TCPNInfo;

// ***DEFINES************************************************************************* 

#define TOPOLOGY_MY_RANK (-0xFFFF)

// ***CLASSES************************************************************************* 

struct TTopologyMutable

{
    TArray< TCPNInfo *> SortedInfos;
    int                 MyRank;
};

// ===================================================================================

/** Base class for all topology classes. It defines a trivial rank of 0 to 
  number of CPN minus 1 by sorting the CPN ID. The master process  usually has rank 0.
  A wrap-around is assumed for the ranks. So -1 is equivalent to MaxRank+1 and vice versa.
  */

class TTopology : public TBaseObject, public TTopologyUser

{
public:

    TTopology(const TDLList<TCPNInfo> *theList);
    virtual ~TTopology(void);

    /** Return CPN info for a given rank. */
    virtual const TCPNInfo *
        GetCPNInfoByRank(int aRank) const;

    TString
        GetClassName(void) const;

    virtual TString
        GetInstanceName(void) const;

    /** Return a list of all CPNs in the topology. */
    const TDLList<TCPNInfo> *
        GetList(void) const;

    /** Return the info for CPN which has a rank by one higher than {\tt aCPNInfo}. 
      If {\tt 0} is passed, the local CPN is assumed as reference. */
    const TCPNInfo *
        GetNext(const TCPNInfo *aCPNInfo = 0) const;

    /** Return the info for CPN which has a rank by one lower than {\tt aCPNInfo}. 
      If {\tt 0} is passed, the local CPN is assumed as reference. */
    const TCPNInfo *
        GetPrev(const TCPNInfo *aCPNInfo = 0) const;

    inline int
        GetPrevRank(int aRank = TOPOLOGY_MY_RANK) const;

    /* Return CPN info which is {\tt Displacement} ranks away for local CPN. 
       Displacement can be positive and negative. */
    const TCPNInfo *
        GetRelative(int Displacement) const;

    inline int
        GetNextRank(int aRank = TOPOLOGY_MY_RANK) const;

    /* Returns the normalized rank between 0 and number of CPN - 1 for any given rank. */
    inline int
        GetNormalizedRank(int aRank) const;

    inline int
        GetNrOfCPN(void) const;

    /** Get rank of CPN. If {\tt 0} is passed, the local CPN is assumed as reference. */
    virtual int
        GetRank(const TCPNInfo *aCPNInfo = 0) const;

    /** Return the type of the topology. There are several child classes for topology. 
      Use this type to cast the topology pointer to the correct class. */
    virtual TTopologyType
        GetType(void) const;

protected:

    virtual bool
        DetermineRank(void) const;

    TTopologyMutable *Mutable;

private:

    const TDLList<TCPNInfo> *List;
};

// ***INLINE METHODS******************************************************************

#ifndef NO_INLINE
#  include "TTopology.i.C"
#endif

// ===================================================================================
// Local Variables:
// mode:c++
// End:
#endif





