/*      file     : parbalance/TDynamicBalancer.C
        project  : Parallel Toolbox 2.x
        author   : Marcus Rickert
        RCS      : $Id: TDynamicBalancer.C,v 1.14 2005/03/29 18:06:45 mr Exp $

        Copyright (C) 2005 Marcus Rickert

        http://www.the-rickerts.de/mr/PAMINA.html
*/

// ***INCLUDES************************************************************************ 

#include <math.h>

#include <ParToolsConfig.h>
#include <TTimer.h>
#include <TStream.h>
#include <TRandom.h>
#include <TEnvironment.h>
#include <TStatistics.h>
#include <TNode.h>
#include <TEdge.h>
#include <TIncidenceInfo.h>
#include <TObjectManager.h>
#include <TNetwork.h>
#include <TCommunication.h>
#include <TTopology.h>
#include <TParallel.h>
#include <TConfiguration.h>
#include <TExtendedCPNInfo.h>
#include <TCommunicationLink.h>
#include <TSamplingDescription.h>
#include <TAggregationInfo.h>
#include <TAggregationResult.h>
#include <TStatHandler.h>
#include <TStatManager.h>

#include "ParBalanceTypes.h"
#include "TMigrationContext.h"
#include "TLoadStat.h"
#include "TExtendedLoadStat.h"
#include "TNeighborCommInfo.h"
#include "TTransferLoadInfo.h"
#include "TLocalBalancer.h"
#include "TNodeHeapEntry.h"

#ifdef USE_PLB
#  include "TPLBalancer.h"
#endif

#include "ParBalanceConfig.h"

#include "TDynamicBalancer.h"

// ***DEFINES************************************************************************* 

#define BALANCER_DEFAULT_USELOCAL               1L
#define BALANCER_DEFAULT_USEPLB                 0L
#define BALANCER_DEFAULT_TRANSFER               0L
#define BALANCER_DEFAULT_GLOBALLOADSTATINTERVAL 10L
#define BALANCER_DEFAULT_GLOBALLOADSTATDUMP     0L

// ***GLOBAL VARIABLES****************************************************************  


TDynamicBalancer *DynamicBalancer = 0;

// ***METHODS************************************************************************* 

// =================================================================================== 
#define ParentClass TBaseObject
// =================================================================================== 

TDynamicBalancer::TDynamicBalancer(TNodeList *theNodes, TEdgeList *theEdges, 
                                   TEdgeList *theBoundaryEdges, 
                                   TTimer *theNodeTimer, TTimer *theEdgeTimer,
                                   TTimer *theAppMiscTimer)

{
    if (DynamicBalancer)
        Error("another instance of class TDynamicBalancer already active");     
    DynamicBalancer = this;
    NeighborCommAggregationInfo = 0;
    LoadAggregationInfo = 0;
    PreviousNumberOfNodes = INVALID;
    Transfer = FALSE;
    LatestGlobalLoadStat = new TLoadStat;
    NodeHeap = 0;
    Topology = 0;
    PerfHistory = 0;
    TimeHistory = 0;
    Nodes = theNodes;
    Edges = theEdges;
    BoundaryEdges = theBoundaryEdges;
    NodeTimer = theNodeTimer;
    EdgeTimer = theEdgeTimer;
    AppMiscTimer = theAppMiscTimer;
    Timer = new TTimer;
    Random= new TRandom;
    Timer->Start();
    Balancer = 0;
    ApplicationTime = MiscTime = AppMiscTime = IdleTime = (double) INVALID;
    ApplicationLoad = (double) INVALID;
    Perf = (double) INVALID;
    LoadStatOutput = 0;
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_GLOBALLOADSTATINTERVAL,
                                           BALANCER_DEFAULT_GLOBALLOADSTATINTERVAL));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_GLOBALLOADSTATDUMP,
                                           BALANCER_DEFAULT_GLOBALLOADSTATDUMP));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_SAMPLINGINTERVAL, 
                                           BALANCER_DEFAULT_SAMPLINGINTERVAL));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_PERFHISTORYLENGTH, 
                                           BALANCER_DEFAULT_PERFHISTORYLENGTH));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_TIMEHISTORYLENGTH, 
                                           BALANCER_DEFAULT_TIMEHISTORYLENGTH));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_TRANSFERDELAY, 
                                           BALANCER_DEFAULT_TRANSFERDELAY));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_TRANSFERDELAYVAR, 
                                           BALANCER_DEFAULT_TRANSFERDELAYVAR));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_USELOCAL, 
                                           BALANCER_DEFAULT_USELOCAL));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_USEPLB, 
                                           BALANCER_DEFAULT_USEPLB));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_TRANSFER, 
                                           BALANCER_DEFAULT_TRANSFER));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_TIMEWEIGHT, 
                                           BALANCER_DEFAULT_TIMEWEIGHT));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_APPLOADWEIGHT, 
                                           BALANCER_DEFAULT_APPLOADWEIGHT));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_LOADWEIGHT, 
                                           BALANCER_DEFAULT_LOADWEIGHT));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_EXCESSLOADWEIGHT, 
                                           BALANCER_DEFAULT_EXCESSLOADWEIGHT));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_LACKOFLOADWEIGHT, 
                                           BALANCER_DEFAULT_LACKOFLOADWEIGHT));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_INCIDENCEWEIGHT, 
                                           BALANCER_DEFAULT_INCIDENCEWEIGHT));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_DISTANCEWEIGHT, 
                                           BALANCER_DEFAULT_DISTANCEWEIGHT));
    Configuration->SetDefault(TConfigEntry(BALANCER_GROUP, BALANCER_NAME_MINPENALTYCHANGE, 
                                           BALANCER_DEFAULT_MINPENALTYCHANGE));
    debug(Print(DM_LIFESPAN, "Run"));
}

// =================================================================================== 

TDynamicBalancer::~TDynamicBalancer(void)

{
    DeleteObject(LoadStatOutput);
    DeleteObject(LatestGlobalLoadStat);
    DeleteObject(Timer);
    DeleteObject(Random);
    DeleteObject(Balancer);
    DeleteObject(PerfHistory);
    DeleteObject(TimeHistory);
    DeleteObject(NodeHeap);
    DynamicBalancer = 0;
}

// =================================================================================== 

bool
TDynamicBalancer::Activate(void)

{
    Print(DM_LIFESPAN, "Activate");
    ObjectManager->RegisterClass(0, TNodeHeapEntry::ClassInfo, CLASS_NODE, 
                                 CLASS_NODE_HEAP_ENTRY, this);
    ObjectManager->RegisterClass(0, TLoadStat::ClassInfo, CLASS_EXTENDED_CPNINFO, 
                                 CLASS_LOAD_STAT, this);
    ObjectManager->RegisterClass(0, TExtendedLoadStat::ClassInfo, CLASS_EXTENDED_CPNINFO, 
                                 CLASS_EXTENDED_LOAD_STAT, this);
    ObjectManager->RegisterClass(0, TTransferLoadInfo::ClassInfo, 
                                 CLASS_EXTENDED_CPNINFO, CLASS_TRANSFER_LOAD_INFO, this);
    ObjectManager->RegisterClass(0, TNeighborCommInfo::ClassInfo,
                                 CLASS_COMMUNICATION_LINK, CLASS_NEIGHBOR_COMM_INFO, this);
    Communication->RegisterRecipient(MTAG_LOAD_BALANCER, this);
    SamplingInterval = (int) Configuration->GetLong(BALANCER_GROUP, 
                                                    BALANCER_NAME_SAMPLINGINTERVAL);
    PerfHistoryLength = (int) Configuration->GetLong(BALANCER_GROUP, 
                                                     BALANCER_NAME_PERFHISTORYLENGTH);
    TimeHistoryLength = (int) Configuration->GetLong(BALANCER_GROUP, 
                                                     BALANCER_NAME_TIMEHISTORYLENGTH);
    TimeWeight = (double) Configuration->GetDouble(BALANCER_GROUP, 
                                                   BALANCER_NAME_TIMEWEIGHT);
    AppLoadWeight = (double) Configuration->GetDouble(BALANCER_GROUP, 
                                                      BALANCER_NAME_APPLOADWEIGHT);
    LoadWeight = (double) Configuration->GetDouble(BALANCER_GROUP, 
                                                   BALANCER_NAME_LOADWEIGHT);
    ExcessLoadWeight = Configuration->GetDouble(BALANCER_GROUP, 
                                                BALANCER_NAME_EXCESSLOADWEIGHT);
    LackOfLoadWeight = Configuration->GetDouble(BALANCER_GROUP, 
                                                BALANCER_NAME_LACKOFLOADWEIGHT);
    IncidenceWeight = Configuration->GetDouble(BALANCER_GROUP, 
                                               BALANCER_NAME_INCIDENCEWEIGHT);
    DistanceWeight = Configuration->GetDouble(BALANCER_GROUP, 
                                              BALANCER_NAME_DISTANCEWEIGHT);
    MinPenaltyChange = Configuration->GetDouble(BALANCER_GROUP, 
                                                BALANCER_NAME_MINPENALTYCHANGE);
    if (ExcessLoadWeight + LackOfLoadWeight + IncidenceWeight + DistanceWeight > 0.0)
        NodeHeap = new THeap< TNodeHeapEntry >;
    PerfHistory = new TStatistics(PerfHistoryLength);
    TimeHistory = new TStatistics(TimeHistoryLength);
    TransferDelay = (int) Configuration->GetLong(BALANCER_GROUP, BALANCER_NAME_TRANSFERDELAY);
    TransferDelayVariation = (int) Configuration->GetLong(BALANCER_GROUP, 
                                                          BALANCER_NAME_TRANSFERDELAYVAR);
    UseLocal = Configuration->GetLong(BALANCER_GROUP, BALANCER_NAME_USELOCAL);
    UsePLB = Configuration->GetLong(BALANCER_GROUP, BALANCER_NAME_USEPLB);
    Transfer = Configuration->GetLong(BALANCER_GROUP, BALANCER_NAME_TRANSFER);
    if (UsePLB)
    {
#ifdef USE_PLB
        Balancer = new TPLBalancer;
#else
        Warning("Activate: PLB not supported");
#endif
    }
    if (Balancer == 0 AND UseLocal)
        Balancer = new TLocalBalancer;
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::ActivateStatistics(void)

{
    if (Parallel->IsMaster())
    {
        LoadStatInterval = (int) Configuration->GetLong(BALANCER_GROUP, 
                                                        BALANCER_NAME_GLOBALLOADSTATINTERVAL);
        LoadStatDump = (int) Configuration->GetLong(BALANCER_GROUP, 
                                                    BALANCER_NAME_GLOBALLOADSTATDUMP);
        
        // start global load statistics:
        
        if (LoadStatInterval > 0)
        {
            LoadSamplingDescription = 
              new TSamplingDescription(TLoadStat::ClassInfo->GetID(), SF_AGGREGATE,
                                       Parallel->GetLocalCPNID(), 
                                       SAMPLING_ID_GLOBAL_LOAD_STAT);
            LoadSamplingDescription->SetCollectInterval(LoadStatInterval);
            LoadAggregationInfo = new TAggregationInfo(this, LoadSamplingDescription);
            StatManager->RegisterAggregation(LoadAggregationInfo);
            StatManager->StartSampling(LoadSamplingDescription);
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::AddBoundaryNode(TMigrationContext &C, TNode *InitialNode)

{
    TInfoID AdjInfoID, LocInfoID;

    AdjInfoID = TAdjacenceInfo::ClassInfo->GetInfoID();
    LocInfoID = TLocationInfo::ClassInfo->GetInfoID();
    if (!C.BoundaryNodes.IsInList(InitialNode))
    {
        TNodeList Cluster;
        TNodeList Pool;
        double    Load;
        TCPNID    LocalCPNID;
        
        LocalCPNID = Parallel->GetLocalCPNID();
        Load = 0.0;
        Cluster.Append(Pool);
        C.BoundaryNodes.Append(InitialNode);
        while (Pool.GetNrOfElements() > 0)
        {
            TNode          *Current;
            TIncidenceInfo *IncInfo;
            int             CEdge;
            TEdge          *Edge;
            
            Current = Pool.GetFirst();
            Cluster.Append(Current);
            Pool.Delete(Current);
            Load += Current->GetLoad(LocalCPNID, AdjInfoID, LocInfoID);
            if ((IncInfo = Current->GetIncidenceInfo()) == 0)
            {
                Error("AddNodeToContext: cannot retrieve incidence info");
                return FALSE;
            }
            for (CEdge = 0; CEdge < IncInfo->GetNrOfIncomingEdges(); CEdge++)
            {
                TNode *Node;
                
                Edge = IncInfo->GetIncomingEdge(CEdge);
                if ((Node = Edge->GetFromNode()) == 0)
                {
                    Error("cannot retrieve FromNode");
                    return FALSE;
                }
                if (Node->IsLocal(LocalCPNID))
                    if (!(Node->IsInList(&C.Nodes) OR
                          Node->IsInList(&Cluster) OR
                          Edge->IsDistributable()))
                    {
                        Cluster.Append(Node);
                        Pool.Append(Node);
                    }
            }
            for (CEdge = 0; CEdge < IncInfo->GetNrOfOutgoingEdges(); CEdge++)
            {
                TNode *Node;
                
                Edge = IncInfo->GetOutgoingEdge(CEdge);
                if ((Node = Edge->GetToNode()) == 0)
                {
                    Error("cannot retrieve ToNode");
                    return FALSE;
                }
                if (Node->IsLocal(LocalCPNID))
                    if (!(Node->IsInList(&C.Nodes) OR
                          Node->IsInList(&Cluster) OR
                          Edge->IsDistributable()))
                    {
                        Cluster.Append(Node);
                        Pool.Append(Node);
                    }
            }
        }
        if (NodeHeap)
        {
            TNodeHeapEntry *HeapEntry;
            TInfoID         LocationInfoID;
            
            if ((HeapEntry = (TNodeHeapEntry *) 
                 InitialNode->GetInfo(TNodeHeapEntry::ClassInfo->GetInfoID(), TRUE)) == 0)
            {
                Error("cannot retrieve heap entry for node %s", 
                      (char *) InitialNode->GetInstanceName());
                return FALSE;
            }
            HeapEntry->Load = Load;
            LocationInfoID = TLocationInfo::ClassInfo->GetInfoID();
            HeapEntry->Distance = 
                MyCenterOfGravity.DistanceTo(InitialNode->Loc(LocationInfoID));
            UpdateHeapEntry(C, HeapEntry);
            NodeHeap->Add(HeapEntry);
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::AddNodeToContext(TMigrationContext &C, TNode *InitialNode)
    
{
    TNodeList Pool;
    TCPNID    LocalCPNID;
    TInfoID   AdjInfoID, LocInfoID;

    AdjInfoID = TAdjacenceInfo::ClassInfo->GetInfoID();
    LocInfoID = TLocationInfo::ClassInfo->GetInfoID();
    LocalCPNID = Parallel->GetLocalCPNID();
    Pool.Append(InitialNode);
    while (Pool.GetNrOfElements() > 0)
    {
        TNode          *Current;
        TIncidenceInfo *IncInfo;
        int             CEdge;
        TEdge          *Edge;
        double          Load;

        Current = Pool.GetFirst();
        C.Nodes.Append(Current);
        Pool.Delete(Current);
        RemoveBoundaryNode(C, Current);
        Load = Current->GetLoad(LocalCPNID, AdjInfoID, LocInfoID);
        C.Load += Load;
        C.RemainingTransferLoad -= Load;
        //Print(DM_BALANCE, "add node %s with load %lf", 
        //      (char *) Current->GetInstanceName(), Load);
        if ((IncInfo = Current->GetIncidenceInfo()) == 0)
        {
            Error("AddNodeToContext: cannot retrieve incidence info");
            return FALSE;
        }
        for (CEdge = 0; CEdge < IncInfo->GetNrOfIncomingEdges(); CEdge++)
        {
            TNode *Node;

            Edge = IncInfo->GetIncomingEdge(CEdge);
            if ((Node = Edge->GetFromNode()) == 0)
            {
                Error("cannot retrieve FromNode");
                return FALSE;
            }
            if (Node->IsLocal(LocalCPNID))
            {
                if (!Node->IsInList(&C.Nodes))
                    if (Edge->IsDistributable())
                    {
                        if (!AddBoundaryNode(C, Node))
                            return FALSE;
                    }
                    else
                        Pool.Append(Node);
            }
            else
            {
                TCPNInfo *CPNInfo;
                
                if ((CPNInfo = Parallel->GetCPNInfoByID(Node->GetCPNID())) == 0)
                {
                    Error("cannot retrieve CPNInfo");
                    return FALSE;
                }
                C.CPNInfos.Append(CPNInfo);
            }
        }
        for (CEdge = 0; CEdge < IncInfo->GetNrOfOutgoingEdges(); CEdge++)
        {
            TNode *Node;

            Edge = IncInfo->GetOutgoingEdge(CEdge);
            if ((Node = Edge->GetToNode()) == 0)
            {
                Error("cannot retrieve ToNode");
                return FALSE;
            }
            if (Node->IsLocal(LocalCPNID))
            {
                if (!Node->IsInList(&C.Nodes))
                    if (Edge->IsDistributable())
                    {
                        if (!AddBoundaryNode(C, Node))
                            return FALSE;
                    }
                    else
                        Pool.Append(Node);
            }
            else
            {
                TCPNInfo *CPNInfo;
                
                if ((CPNInfo = Parallel->GetCPNInfoByID(Node->GetCPNID())) == 0)
                {
                    Error("cannot retrieve CPNInfo");
                    return FALSE;
                }
                C.CPNInfos.Append(CPNInfo);
            }
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::CancelSend(TTransferLoadInfo *Info, const TTimeStep *TimeStep)

{
    if (TimeStep)
        Print(DM_BALANCE, "in time-step %s: cancel %s", (char *) TimeStep->GetInstanceName(),
              (char *) Info->GetInstanceName());
    else
        Print(DM_BALANCE, "cancel %s", (char *) Info->GetInstanceName());       
    if (Info->GetMode() == SendLoad)
    {
        if (!SendTimedEvent(TDestination(&Info->GetMigrationContext()->CPNInfos,
                                         DEST_FLAG_EXCLUDE_MYSELF), 
                            MEVENT_BALANCE_CANCEL_SEND, Info->GetTransferTimeStep()))
        {
            Error("CancelSend: cannot send event");
            return FALSE;
        }
    }
    Info->Cancel();
    PendingTransfers.Delete(Info);
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::CheckLocalStats(void)

{
    if (PreviousNumberOfNodes != Nodes->GetNrOfElements())
    {
        TNodeEnum  CNode;
        TInfoID    LocationInfoID;
        
        LocationInfoID = TLocationInfo::ClassInfo->GetInfoID();
        MyCenterOfGravity.Reset();
        MyFrame.Empty();
        CNode.WalksForwardThrough(Nodes);
        while (CNode())
        {
            MyCenterOfGravity += CNode->Loc(LocationInfoID);
            MyFrame.Merge(CNode->Loc(LocationInfoID));
            CNode++;
        }
        if (Nodes->GetNrOfElements())
            MyCenterOfGravity /= (double) Nodes->GetNrOfElements();
        MyDiameter = sqrt(MyFrame.GetDeltaX()*MyFrame.GetDeltaX() +
                          MyFrame.GetDeltaY()*MyFrame.GetDeltaY());
        PreviousNumberOfNodes = Nodes->GetNrOfElements();
    }
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::CheckPendingTransfers(TTimeStep &TimeStep)

{
    TTransferLoadInfoEnum CInfo;

    CInfo.WalksForwardThrough(PendingTransfers);
    while (CInfo())
    {
        if (TimeStep > CInfo->GetTransferTimeStep())
        {
            if (!CancelSend(CInfo, &TimeStep))
            {
                Error("CheckPendingTransfers: cannot cancel");
                return FALSE;
            }
        }
        else if (CInfo->GetTransferTimeStep() == TimeStep)
        {
            switch (CInfo->GetMode())
            {
            case WaitForAcknowledge:
                
                if (CInfo->GetNrOfMissingAcks() == 0)
                {
                    Print(DM_CONTROL, "execute %s in time-step %s",
                          (char *) CInfo->GetInstanceName(),
                          (char *) TimeStep.GetInstanceName());
                    if (!Migrate(CInfo))
                    {
                        Error("CheckPendingTransfers: cannot migrate");                 
                        return FALSE;
                    }
                    CInfo->Reset();
                    PendingTransfers.Delete(CInfo);
                }
                else
                    if (!CancelSend(CInfo, &TimeStep))
                    {
                        Error("CheckPendingTransfers: cannot cancel");
                        return FALSE;
                    }
                break;

            case WaitForConfirm:

                Print(DM_COLLISION, "COLLISION: %s not confirmed -> cancelling",
                      (char *) CInfo->GetInstanceName());
                CInfo->Reset();
                PendingTransfers.Delete(CInfo);
                break;

            case ReceiveLoad:

                if (!Wait(CInfo))
                {
                    Error("CheckPendingTransfers: while waiting for %s",
                          (char *) CInfo->GetInstanceName());
                    return FALSE;
                }
                CInfo->Reset();
                PendingTransfers.Delete(CInfo);
                break;

            default:
                ;
            }
        }
        CInfo++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::CheckNewTransfers(TTimeStep &TimeStep)

{
    TEnumerator< TExtendedCPNInfo > 
        CInfo(TExtendedCPNInfo::ClassInfo->GetTable());
    TInfoID InfoID;
    double MaxLoad;
    TTransferLoadInfo *MaxInfo;

    InfoID = TTransferLoadInfo::ClassInfo->GetInfoID();
    MaxLoad = 0;
    MaxInfo = 0;
    CInfo.Enumerate();
    while (CInfo())
    {
        TTransferLoadInfo *Info;

        if ((Info = (TTransferLoadInfo *) CInfo->GetInfo(InfoID)))
        {
            if (Info->GetMode() == TransferSelected AND
                Info->GetTransferLoad() > MaxLoad)
            {
                MaxLoad = Info->GetTransferLoad();
                MaxInfo = Info;
                Print(DM_LOAD, "pending in time-step %s: %s", 
                      (char *) TimeStep.GetInstanceName(), 
                      (char *) Info->GetObjectName());      
            }
        }
        CInfo++;
    }
    if (MaxInfo AND !MaxInfo->IsInList(&PendingTransfers))
        if (!PrepareTransfer(TimeStep, MaxInfo))
        {
            Error("CheckNewTransfers: cannot prepare transfer");
            return FALSE;
        }
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::Collect(TSamplingDescription *Description)

{
    if (Description->GetStatObjectClassID() == TLoadStat::ClassInfo->GetID())
    {
        TLoadStat *Stat = new TLoadStat(Parallel->GetLocalCPNID());

        Stat->SetTimeStep(Description->GetTimeStep());
        Stat->Collect(SamplingInterval, IdleTime, MiscTime, 
                      ApplicationTime, AppMiscTime, ApplicationLoad, Perf);
        Description->GetObjectList()->Append(Stat);
        return TRUE;
    }
    if (Description->GetStatObjectClassID() == TExtendedLoadStat::ClassInfo->GetID())
    {
        TExtendedLoadStat *Stat = new TExtendedLoadStat(Parallel->GetLocalCPNID());

        Stat->SetTimeStep(Description->GetTimeStep());
        Stat->Collect(SamplingInterval, IdleTime, MiscTime, 
                      ApplicationTime, AppMiscTime, ApplicationLoad, Perf);
        Stat->ExtendedCollect(MyCenterOfGravity, MyFrame);
        Description->GetObjectList()->Append(Stat);
        return TRUE;
    }
    if (Description->GetStatObjectClassID() == TNeighborCommInfo::ClassInfo->GetID())

        return CollectNeighborCommInfos(Description->GetObjectList());

    return TStatHandler::Collect(Description);
}

// =================================================================================== 

bool
TDynamicBalancer::CollectNeighborCommInfos(TStatObjectList *StatList)

{
    TEnumerator< TExtendedCPNInfo > CInfo(TExtendedCPNInfo::ClassInfo->GetTable());
    TNeighborCommInfoList *InfoList = (TNeighborCommInfoList *) StatList;

    CInfo.Enumerate();
    while (CInfo())
    {
        if (CInfo->CPNInfo->ID != Parallel->GetLocalCPNID())
        {
            TNeighborCommInfo *Info;
            TParallelID        LinkID;

            LinkID = Network->GetCommunicationLinkID(Parallel->GetLocalCPNID(), 
                                                     CInfo->CPNInfo->ID);
            Info = new TNeighborCommInfo(LinkID);
            Info->Collect(CInfo);           
            InfoList->Append(Info);
        }
        CInfo++;
    }
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::ComputeLocalLoadValues(void)

{
    TNodeEnum CNode;    
    TEdgeEnum CEdge;
    double    NewApplicationTime;
    TInfoID   AdjInfoID, LocInfoID;

    AdjInfoID = TAdjacenceInfo::ClassInfo->GetInfoID();
    LocInfoID = TLocationInfo::ClassInfo->GetInfoID();    
    IdleTime = Communication->GetIdleTimer()->GetLap();
    NewApplicationTime = NodeTimer->GetLap() + EdgeTimer->GetLap();
    AppMiscTime = AppMiscTimer->GetLap();
    MiscTime = Timer->GetLap() - (IdleTime + ApplicationTime + AppMiscTime);

    NodeLoad = 0.0;
    CNode.WalksForwardThrough(Nodes);
    while (CNode())
    {
        NodeLoad += CNode->GetLoad(Parallel->GetLocalCPNID(), INVALID, LocInfoID);
        CNode++;
    }    
    EdgeLoad = 0.0;
    CEdge.WalksForwardThrough(Edges);
    while (CEdge())
    {
        EdgeLoad += CEdge->GetLoad(Parallel->GetLocalCPNID(), AdjInfoID, LocInfoID);
        CEdge++;
    }    
    ApplicationLoad = NodeLoad + EdgeLoad;
    if (NewApplicationTime > 0.0)
    {
        if (TimeHistory)
        {
            TimeHistory->AddValue(NewApplicationTime);
            ApplicationTime = TimeHistory->GetMean();
        }
        if (PerfHistory)
        {
            PerfHistory->AddValue(ApplicationLoad / NewApplicationTime);
            Perf = PerfHistory->GetMean();
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::DoubleCheckMigration(TTransferLoadInfo *Info)

{
    TMigrationContext    *C;
    TNodeEnum             CNode;
    TCPNID                LocalCPNID;
    TNodeList             IndependantNodes;
        
    LocalCPNID = Parallel->GetLocalCPNID();
    C = Info->GetMigrationContext();
    CNode.WalksForwardThrough(C->Nodes);
    while (CNode())
    {
        if (CNode->IsLocal(LocalCPNID))
        {
            TIncidenceInfo *IncInfo;
            int             CEdge;

            if ((IncInfo = CNode->GetIncidenceInfo()) == 0)
            {
                Error("DoubleCheckMigration: cannot retrieve incidence info");
                return FALSE;
            }
            for (CEdge = 0; CEdge < IncInfo->GetNrOfIncomingEdges(); CEdge++)
            {
                TNode *Node;
                TEdge *Edge;

                Edge = IncInfo->GetIncomingEdge(CEdge);
                if ((Node = Edge->GetFromNode()) == 0)
                {
                    Error("cannot retrieve FromNode");
                    return FALSE;
                }
                if (!Node->IsLocal(LocalCPNID))
                {
                    TCPNInfo *CPNInfo;
                    
                    if ((CPNInfo = Parallel->GetCPNInfoByID(Node->GetCPNID())) == 0)
                    {
                        Error("cannot retrieve CPNInfo");
                        return FALSE;
                    }
                    if (!C->CPNInfos.IsInList(CPNInfo))
                    {
                        Warning("node %s is on independant CPN %lX",
                                (char *) Node->GetInstanceName(), CPNInfo->ID);
                        IndependantNodes.Append(CNode);
                    }
                }
            }
            for (CEdge = 0; CEdge < IncInfo->GetNrOfOutgoingEdges(); CEdge++)
            {
                TNode *Node;
                TEdge *Edge;

                Edge = IncInfo->GetOutgoingEdge(CEdge);
                if ((Node = Edge->GetToNode()) == 0)
                {
                    Error("cannot retrieve ToNode");
                    return FALSE;
                }
                if (!Node->IsLocal(LocalCPNID))
                {
                    TCPNInfo *CPNInfo;
                    
                    if ((CPNInfo = Parallel->GetCPNInfoByID(Node->GetCPNID())) == 0)
                    {
                        Error("cannot retrieve CPNInfo");
                        return FALSE;
                    }
                    if (!C->CPNInfos.IsInList(CPNInfo))
                    {
                        Warning("node %s is on independant CPN %lX",
                                (char *) Node->GetInstanceName(), CPNInfo->ID);
                        IndependantNodes.Append(CNode);
                    }
                }
            }
        }
        else
        {
            Warning("node %s no longer local", (char *) CNode->GetInstanceName());
            IndependantNodes.Append(CNode);
        }
        CNode++;            
    }
    if (IndependantNodes.GetNrOfElements() > 0)
    {
        CNode.WalksForwardThrough(IndependantNodes);
        while (CNode())
        {
            C->Nodes.Delete(CNode);
            CNode++;
        }       
        Print("%d node(s) removed from migration selection", 
              IndependantNodes.GetNrOfElements());
    }
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::ExecuteSubTimeStep(TTimeStep &TimeStep)

{
    int OffsetInInterval;


    CheckLocalStats();
    if (TimeStep.GetSubStep() == 0)
    {
        if (Transfer)
            if (!CheckNewTransfers(TimeStep))
            {
                Error("ExecuteSubTimeStep: while looking for new transfers");
                return FALSE;
            }
        if (!CheckPendingTransfers(TimeStep))
        {
            Error("ExecuteSubTimeStep: while checking pending transfers");
            return FALSE;
        }
        OffsetInInterval = (int) (TimeStep % SamplingInterval);
    
        if (OffsetInInterval == 0)
        {
            ComputeLocalLoadValues();
            if (Balancer)
                Balancer->SetLocalLoadValues(ApplicationTime, ApplicationLoad, Perf);
        }
    }
    if (Balancer)
    {
        if (!Balancer->ExecuteSubTimeStep(TimeStep))
        {
            Error("ExecuteSubTimeStep: while calling balancer");            
            return FALSE;
        }
    }
    return TRUE;
}

// =================================================================================== 

TString
TDynamicBalancer::GetClassName(void) const

{
    return "TDynamicBalancer";
}

// =================================================================================== 

TLoadStat *
TDynamicBalancer::GetLatestGlobalLoadStat(void) 

{
    return LatestGlobalLoadStat;
}

// =================================================================================== 

bool
TDynamicBalancer::HandleAggregationResult(TAggregationInfo *AggregationInfo, 
                                          TTimeStep &TimeStep)

{
    if (AggregationInfo == NeighborCommAggregationInfo)
        return HandleNeighborCommResult(AggregationInfo, TimeStep);
    if (AggregationInfo == LoadAggregationInfo)
        return HandleLoadStatResult(AggregationInfo, TimeStep);
    return TStatHandler::HandleAggregationResult(AggregationInfo, TimeStep);
}

// =================================================================================== 

bool
TDynamicBalancer::HandleNeighborCommResult(TAggregationInfo *AggregationInfo, 
                                           TTimeStep &TimeStep)

{
    TAggregationResult    *Result = 0;
    TNeighborCommInfoList *StatList = 0;
    TNeighborCommInfoEnum  CInfo;
    TInfoID                InfoID;

    if ((Result = (TAggregationResult *) AggregationInfo->GetResult(TimeStep)) == 0)
    {
        Error("HandleNeighborCommResult: cannot retrieve aggregation result");
        return FALSE;
    }
    if ((StatList = (TNeighborCommInfoList *) Result->StatObjectList) == 0)
    {
        Error("HandleNeighborCommResult: cannot retrieve stat list");
        return FALSE;
    }
    InfoID = TNeighborCommInfo::ClassInfo->GetInfoID();
    CInfo.WalksForwardThrough(StatList);
    while (CInfo())
    {
        TCommunicationLink *Link;

        if ((Link = Network->GetCommunicationLink(CInfo->GetParallelID(), TRUE)) == 0)
        {
            Error("cannot retrieve link for %ld", CInfo->GetParallelID());
            return FALSE;
        }
        Link->SetInfo(InfoID, (TNeighborCommInfo *) CInfo, TRUE);
        if (CInfo->NrOfBoundaries > 0)
            Link->Activate();
        else
            Link->Deactivate();
        CInfo++;
    }
    AggregationInfo->DisposeResult(TimeStep);
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::HandleCreateRequest(TParallelObject *&Object, 
                                      const TClassInfo *ClassInfo, TParallelID ParallelID)

{
    if (ClassInfo == TLoadStat::ClassInfo)

        Object = new TLoadStat(ParallelID);

    else if (ClassInfo == TExtendedLoadStat::ClassInfo)

        Object = new TExtendedLoadStat(ParallelID);

    else if (ClassInfo == TTransferLoadInfo::ClassInfo)
        
        Object = new TTransferLoadInfo(ParallelID);

    else if (ClassInfo == TNeighborCommInfo::ClassInfo)
        
        Object = new TNeighborCommInfo(ParallelID);

    else if (ClassInfo == TNodeHeapEntry::ClassInfo)
        
        Object = new TNodeHeapEntry(ParallelID);

    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::HandleEvent(TEventID EventID, TBaseMessage *Message, int &Handled)

{
    switch(EventID)
    {

    case MEVENT_BALANCE_REQUEST_SEND:
        Handled = TRUE;
        return HandleEventRequestSend(Message);
        break;

    case MEVENT_BALANCE_ACK_SEND:
        Handled = TRUE;
        return HandleEventAckSend(Message);
        break;

    case MEVENT_BALANCE_READY:
        Handled = TRUE;
        return HandleEventReady(Message, Handled);
        break;

    case MEVENT_BALANCE_DENY_SEND:
        Handled = TRUE;
        return HandleEventDenySend(Message);
        break;

    case MEVENT_BALANCE_CANCEL_SEND:
        Handled = TRUE;
        return HandleEventCancelSend(Message);
        break;

    case MEVENT_BALANCE_CONFIRM_SEND:
        Handled = TRUE;
        return HandleEventConfirmSend(Message);
        break;

    default:
        Error("illegal event ID %ld", EventID);
        return FALSE;
    }
#ifndef COMPLETE_SWITCH_FLOW
    return TRUE;
#endif
}


// =================================================================================== 

bool
TDynamicBalancer::HandleEventAckSend(TBaseMessage *Message)

{
    TTimeStep          RefTimeStep;
    TTransferLoadInfo *Info;

    if (!HandleTimedEvent(Message, Info, RefTimeStep))
    {
        Error("HandleEventAckRequest: cannot decode");
        return FALSE;
    }
    if (Info)
    {
        Info->Acknowledge();
        if (Info->GetNrOfMissingAcks() == 0)
            if (!SendTimedEvent(TDestination(&Info->GetMigrationContext()->CPNInfos,
                                             DEST_FLAG_EXCLUDE_MYSELF), 
                                MEVENT_BALANCE_CONFIRM_SEND, 
                                Info->GetTransferTimeStep()))
            {
                Error("HandleEventAckSend: cannot send message");
                return FALSE;
            }
        Print(DM_BALANCE, "received ack for %s from %lX", 
              (char *) Info->GetInstanceName(), Message->GetSender());

    }    
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::HandleEventCancelSend(TBaseMessage *Message)

{
    TTransferLoadInfo *Info;
    TTimeStep          RefTimeStep;

    if (!HandleTimedEvent(Message, Info, RefTimeStep))
    {
        Error("HandleEventCancelSend: cannot decode");
        return FALSE;
    }
    if (Info AND Info->GetParallelID() == Message->GetSender())
    {
        Print(DM_COLLISION, "received cancellation for %s from %lX", 
              (char *) Info->GetInstanceName(), Message->GetSender());
        Info->Cancel();
        PendingTransfers.Delete(Info);
    }    
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::HandleEventConfirmSend(TBaseMessage *Message)

{
    TTransferLoadInfo *Info;
    TTimeStep          RefTimeStep;

    if (!HandleTimedEvent(Message, Info, RefTimeStep))
    {
        Error("HandleEventCancelSend: cannot decode");
        return FALSE;
    }
    if (Info AND Info->GetParallelID() == Message->GetSender())
    {
        Print(DM_BALANCE, "received confirm for %s from %lX",
              (char *) Info->GetInstanceName(), Message->GetSender());
        Info->Confirm();
    }    
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::HandleEventDenySend(TBaseMessage *Message)

{
    TTransferLoadInfo *Info;
    TTimeStep          RefTimeStep;

    if (!HandleTimedEvent(Message, Info, RefTimeStep))
    {
        Error("HandleEventDenySend: cannot decode");
        return FALSE;
    }
    if (Info)
    {   
        Print(DM_COLLISION, "COLLISION: %s denied", (char *) Info->GetInstanceName());
        if (!CancelSend(Info))
        {
            Error("HandleEventDenySend: cannot cancel");
            return FALSE;
        }
        PendingTransfers.Delete(Info);
    }    
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::HandleEventReady(TBaseMessage *Message, int &Handled)

{
    TTimeStep          RefTimeStep;
    TTransferLoadInfo *Info;

    if (!HandleTimedEvent(Message, Info, RefTimeStep))
    {
        Error("HandleEventReady: cannot decode");
        return FALSE;
    }
    if (Info)
    {
        Info->Ready();
        if (Info->GetNrOfNonReadyCPN() == 0)
            Handled = EXIT_LOOP;
        Print(DM_BALANCE, "received ready for %s from %lX", 
              (char *) Info->GetInstanceName(), Message->GetSender());
    }
    else
        Warning("received READY from %lX for non-existing transfer (TS %s)",
                Message->GetSender(), (char *) RefTimeStep.GetInstanceName());
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::HandleEventRequestSend(TBaseMessage *Message)

{
    TTransferLoadInfo *Info;
    TTransferLoadInfo *NewInfo;
    TTimeStep          RefTimeStep;

    if (!HandleTimedEvent(Message, Info, RefTimeStep))
    {
        Error("HandleEventRequestSend: cannot decode");
        return FALSE;
    }
    if (Info)
    {

        Print(DM_COLLISION, "COLLISION: deny transfer from %lX (reason: %s)", 
              Message->GetSender(), (char*) Info->GetInstanceName());

        return SendTimedEvent(Message->GetSender(), MEVENT_BALANCE_DENY_SEND,
                              Info->GetTransferTimeStep());
    }    
    NewInfo = new TTransferLoadInfo(Message->GetSender());
    NewInfo->SetReceiveTimeStep(RefTimeStep);
    PendingTransfers.Append(NewInfo);
    
    Print(DM_BALANCE, "permit %s from %lX", (char *) NewInfo->GetInstanceName(), 
          Message->GetSender());

    return SendTimedEvent(Message->GetSender(), MEVENT_BALANCE_ACK_SEND, 
                          NewInfo->GetTransferTimeStep());
}

// =================================================================================== 

bool
TDynamicBalancer::HandleLoadRequest(TParallelObject *Object, bool &Loaded)

{
    if (Object->GetClassID() == TLoadStat::ClassInfo->GetID())

        Loaded = TRUE;

    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::HandleLoadStatResult(TAggregationInfo *Info, TTimeStep &TimeStep)

{
    TAggregationResult *Result = 0;
    TLoadStat          *Stat = 0;

    if ((Result = (TAggregationResult *) Info->GetResult(TimeStep)) == 0)
    {
        Error("HandleLoadStatResult: cannot retrieve aggregation result");
        return FALSE;
    }
    if ((Stat = (TLoadStat *) Result->StatObject) == 0)
    {
        Error("HandleLoadStatResult: cannot retrieve stat object");
        return FALSE;
    }   
    if (!WriteLoadStat(Stat))
        return FALSE;
    if (LoadStatDump)
        Stat->PrintScreenStat(stderr);
    *LatestGlobalLoadStat = *Stat;
    Info->DisposeResult(TimeStep);
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::HandleTimedEvent(TBaseMessage *Message, TTransferLoadInfo *&Info,
                                   TTimeStep &TimeStep)

{
    TTransferLoadInfoEnum CInfo;

    Info = 0;
    if (!TimeStep.Decode(Message))
    {
        Error("HandleTimedEvent: cannot decode");
        return FALSE;
    }
    CInfo.WalksForwardThrough(PendingTransfers);
    while (CInfo())
    {
        if (CInfo->GetTransferTimeStep() == TimeStep)
        {
            Info = CInfo;
            return TRUE;
        }
        CInfo++;
    }    
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::Migrate(TTransferLoadInfo *Info)

{
    TMigrationContext    *C;
    TCPNInfo             *CPNInfo;
    TCPNInfoEnum          CCPNInfo;
    TNodeEnum             CNode;
    
    CPNInfo = Parallel->GetCPNInfoByID(Info->GetParallelID());
    C = Info->GetMigrationContext();

    // !!!!! we should only check if neccessary

    if (!DoubleCheckMigration(Info))
    {
        Error("Migrate: while double-checking migration selection");
        return FALSE;
    }
    if (!WaitUntilReady(Info))
    {
        Error("Migrate: while waiting for ready events");
        return FALSE;
    }

    Print(DM_MIGRATION, "***** migrate %f load in %d node(s) for %s", C->Load,
          C->Nodes.GetNrOfElements(), (char *) Info->GetInstanceName());

    CCPNInfo.WalksForwardThrough(C->CPNInfos);
    while (CCPNInfo())
    {
        Print(DM_BALANCE, "neighbouring CPN: %lX", CCPNInfo->ID);
        CCPNInfo++;
    }
    if (!(ObjectManager->MigrateContainerObjects(CPNInfo,
                                                 (TContainerObjectList *) &C->Nodes) AND
          ObjectManager->Migrate(TRUE)))
    {
        Error("Error: while migrating nodes");
        return FALSE;
    }                                         
    Print(DM_BALANCE, "finished migrating");
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::NewContext(TTransferLoadInfo *Info, TMigrationContext *&C)

{
    TEdgeEnum CEdge; 
    TExtendedCPNInfo *ECPNInfo;

    if ((ECPNInfo = Network->GetExtendedCPNInfo(Info->GetParallelID(), TRUE)) == 0)
    {
        Error("NewContext: cannot retrieve ExtendedCPNInfo");
        return FALSE;
    }
    if (NodeHeap)
        NodeHeap->DeleteAll(); 
    C = new TMigrationContext(ECPNInfo->BoundaryContainers.GetNrOfElements(),
                              Info->GetTransferLoad());
    C->CPNInfos.Append(ECPNInfo->CPNInfo);
    CEdge.WalksForwardThrough((TEdgeList &) ECPNInfo->BoundaryContainers);
    while (CEdge())
    {
        TNode *Node;

        if ((Node = CEdge->GetFromNode()) == 0)
        {
            Error("cannot retrieve FromNode");
            return FALSE;
        }
        if (Node->IsLocal(Parallel->GetLocalCPNID()))
            if (!AddBoundaryNode(*C, Node))
                return FALSE;
        if ((Node = CEdge->GetToNode()) == 0)
        {
            Error("cannot retrieve ToNode");
            return FALSE;
        }
        if (Node->IsLocal(Parallel->GetLocalCPNID()))
            if (!AddBoundaryNode(*C, Node))
                return FALSE;
        CEdge++;
    }
    if (C->BoundaryNodes.GetNrOfElements() == 0)
        Warning("NewContext: no common nodes with %lX found", ECPNInfo->CPNInfo->ID);
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::WriteLoadStat(TLoadStat *Stat)

{
    if (LoadStatOutput == 0 AND
        Configuration->Exists(BALANCER_GROUP, BALANCER_NAME_GLOBALLOADSTATFILENAME))
    {
        TString Filename;
        
        LoadStatOutput = new TStream;
        Filename = Environment->ReplacePathTokens(Configuration->GetString());
        if (!LoadStatOutput->Open(Filename, "w"))
        {
            Error("ActivateStatistics: cannot open load stat file %s", 
                  Filename());
            return FALSE;
        }
        TLoadStat::WriteStatDescription(LoadStatOutput);
    }
    if (LoadStatOutput)
        if (Stat->GetAveragePerf() != (double) INVALID)
            Stat->Write(0, LoadStatOutput);
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::PrepareTransfer(TTimeStep &TimeStep, TTransferLoadInfo *Info)

{
    TMigrationContext *C = 0;
    TTimeStep          SendTimeStep;

    
    if (!NewContext(Info, C))
    {
        Error("PrepareTransfer: cannot get new migration context");
        return FALSE;
    }
    if (!SelectNodes(Info, *C))
    {
        Error("PrepareTransfer: cannot fill migration context");
        return FALSE;
    }
    SendTimeStep = TimeStep + (TransferDelay + Random->Int(TransferDelayVariation));
    Info->SetSendTimeStep(SendTimeStep, C);
    if (!SendTimedEvent(TDestination(&Info->GetMigrationContext()->CPNInfos,
                                     DEST_FLAG_EXCLUDE_MYSELF), 
                        MEVENT_BALANCE_REQUEST_SEND, SendTimeStep))
    {
        Error("PrepareTransfer: cannot send message");
        return FALSE;
    }
    PendingTransfers.Append(Info);
    Print(DM_BALANCE, "in time-step %s: prepare %s", (char *) TimeStep.GetInstanceName(),
          (char *) Info->GetInstanceName());
    return TRUE;
}

// =================================================================================== 

void
TDynamicBalancer::RemoveBoundaryNode(TMigrationContext &C, TNode *Node)

{
    C.BoundaryNodes.Delete(Node);
}

// =================================================================================== 

bool
TDynamicBalancer::Run(void)

{
    debug(Print(DM_LIFESPAN, "Run"));
    if (!(StatManager->RegisterSamplingHandler(TLoadStat::ClassInfo->GetID(), this) AND
          StatManager->RegisterSamplingHandler(TExtendedLoadStat::ClassInfo->GetID(), this) AND
          StatManager->RegisterSamplingHandler(TNeighborCommInfo::ClassInfo->GetID(), this)))
    {
        Error("Run: cannot register myself as sampling handler");
        return FALSE;
    }
    CheckLocalStats();
    Topology = Parallel->GetPhysicalTopology();
    Random->SetSeed(Parallel->GetLocalCPNID());
    if (!ActivateStatistics())
    {
        Error("Run: cannot start statistics");
        return FALSE;
    }
    if (Balancer)
        if (!Balancer->Run())
        {
            Error("Run: while calling balancer");
            return FALSE;
        }
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::SelectNodes(TTransferLoadInfo *Info, TMigrationContext &C)

{
    while (C.Load < Info->GetTransferLoad())
    {
        TNode *Node;
        
        if (NodeHeap)
        {
            TNodeHeapEntry *HeapEntry;

            UpdateNodeHeap(C);
            if ((HeapEntry = NodeHeap->GetTop()) == 0)
            {
                Error("SelectNodes: empty boundary node heap");
                return FALSE;
            }
            Node = (TNode *) HeapEntry->GetContainer();
        }
        else
            Node = C.BoundaryNodes.GetFirst();
        if (Node == 0)
        {
            Error("SelectNodes: cannot select best node");
            return FALSE;
        }
        if (!AddNodeToContext(C, Node))
        {
            Error("SelectNodes: cannot add node %s", (char *) Node->GetInstanceName());
            return FALSE;
        }
    }
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::SendTimedEvent(TDestination Destination, TEventID EventID,
                                 const TTimeStep &TimeStep)

{
    TBaseMessage *Message;
    
    if (!(Communication->GetNewMessage(Message, Destination,
                                       MTAG_LOAD_BALANCER, EventID) AND
          Message->Event(EventID) AND
          TimeStep.Encode(Message) AND
          Communication->Commit(Message)))
    {
        Error("SendTimedEvent: cannot send timed (%s) event %ld to %s ",
              (char *) TimeStep.GetInstanceName(), EventID, 
              (char *) Destination.GetInstanceName());
        return FALSE;
    }
    return TRUE;
}

// =================================================================================== 

bool
TDynamicBalancer::StartNeighborCommStat(void)

{
    NeighborCommSamplingDescription = 
        new TSamplingDescription(TNeighborCommInfo::ClassInfo->GetID(), 0,
                                 DEST_PARENT, SAMPLING_ID_NEIGHBOR_COMM_STAT);
    NeighborCommSamplingDescription->SetCollectInterval(5);
    NeighborCommAggregationInfo = new TAggregationInfo(this, NeighborCommSamplingDescription);
    StatManager->RegisterAggregation(NeighborCommAggregationInfo);
    StatManager->StartSampling(NeighborCommSamplingDescription);
    return TRUE;
}

// =================================================================================== 

void
TDynamicBalancer::UpdateHeapEntry(TMigrationContext &C, TNodeHeapEntry *HeapEntry)

{
    double Penalty;
    double DistancePenalty;
    double IncidencePenalty;
    double ExcessLoadPenalty;
    double LackOfLoadPenalty;

    if (HeapEntry->Load > C.RemainingTransferLoad)
    {
        ExcessLoadPenalty = ExcessLoadWeight * 
            (HeapEntry->Load - C.RemainingTransferLoad) / ApplicationLoad;
        LackOfLoadPenalty = 0.0;
    }
    else
    {
        LackOfLoadPenalty = LackOfLoadWeight * 
            (C.RemainingTransferLoad - HeapEntry->Load) / ApplicationLoad;
        ExcessLoadPenalty = 0.0;
    }
    
    // !!!!! MISSING: INCIDENCE PENALTY !!!!!

    IncidencePenalty = 0.0;
    
    DistancePenalty =  - DistanceWeight * HeapEntry->Distance / MyDiameter;

    Penalty = ExcessLoadPenalty + LackOfLoadPenalty + IncidencePenalty + DistancePenalty;

    /*
    Print("UpdateHeapEntry: Old:%.3f Sum:%.3f (ExcessLoad:%.3f LackOfLoad:%.3f "
          "Incidence:%.3f Distance:%.3f)", HeapEntry->Penalty, Penalty,
          ExcessLoadPenalty, LackOfLoadPenalty, IncidencePenalty, DistancePenalty);
          */

    if (HeapEntry->Penalty == (double) INVALID OR Penalty < 1.0E-10 OR
        fabs(Penalty - HeapEntry->Penalty) / HeapEntry->Penalty > MinPenaltyChange)
        HeapEntry->Penalty = Penalty;

}

// =================================================================================== 

void
TDynamicBalancer::UpdateNodeHeap(TMigrationContext &C)

{
    int CHeap;

    for (CHeap = 0; CHeap < NodeHeap->GetNrOfElements(); CHeap++)
        UpdateHeapEntry(C, NodeHeap->Get(CHeap));
}

// =================================================================================== 
    
bool
TDynamicBalancer::Wait(TTransferLoadInfo *Info)

{
    Print(DM_CONTROL, "waiting for %s", (char *) Info->GetInstanceName());
    if (!SendTimedEvent(Info->GetParallelID(), MEVENT_BALANCE_READY, 
                        Info->GetTransferTimeStep()))
    {
        Error("cannot send ready event");
        return FALSE;
    }
    return ObjectManager->Wait(Info->GetParallelID());
}

// =================================================================================== 

bool
TDynamicBalancer::WaitUntilReady(TTransferLoadInfo *Info)

{
    Info->SetMode(WaitUntilCPNReady);
    Print(DM_MIGRATION, "waiting for ready events for %s",
          (char *) Info->GetInstanceName());
    while (Info->GetNrOfNonReadyCPN() > 0)
        if (!Communication->MessageLoop(TRUE))
        {
            Error("WaitUntilReady: while waiting for ready events");
            return FALSE;
        }
    Info->SetMode(SendLoad);
    return TRUE;
}

// =================================================================================== 
#undef ParentClass



